@@ -1,7 +1,7 @@
 /*
  * Please refer to the documentation hosted at
  *
- *   https://docs.rs/blazesym-c/0.1.1
+ *   https://docs.rs/blazesym-c/0.1.4
  */
 
 
@@ -64,7 +64,8 @@ typedef int16_t blaze_err;
 #define BLAZE_ERR_INVALID_INPUT -256
 /**
  * An error returned when an operation could not be completed
- * because a call to [`write`] returned [`Ok(0)`].
+ * because a call to [`write`][std::io::Write::write] returned
+ * [`Ok(0)`][Ok].
  */
 #define BLAZE_ERR_WRITE_ZERO -257
 /**
@@ -264,6 +265,18 @@ typedef uint8_t blaze_normalize_reason;
  * The address belonged to an entity that is currently unsupported.
  */
 #define BLAZE_NORMALIZE_REASON_UNSUPPORTED 2
+/**
+ * The file offset does not map to a valid piece of code/data.
+ */
+#define BLAZE_NORMALIZE_REASON_INVALID_FILE_OFFSET 3
+/**
+ * The symbolization source has no or no relevant symbols.
+ */
+#define BLAZE_NORMALIZE_REASON_MISSING_SYMS 4
+/**
+ * The address could not be found in the symbolization source.
+ */
+#define BLAZE_NORMALIZE_REASON_UNKNOWN_ADDR 5
 
 /**
  * The valid variant kind in [`blaze_user_meta`].
@@ -272,14 +285,30 @@ typedef uint8_t blaze_user_meta_kind;
 /**
  * [`blaze_user_meta_variant::unknown`] is valid.
  */
-#define BLAZE_USER_META_KIND_BLAZE_USER_META_UNKNOWN 0
+#define BLAZE_USER_META_KIND_UNKNOWN 0
 /**
  * [`blaze_user_meta_variant::apk`] is valid.
  */
-#define BLAZE_USER_META_KIND_BLAZE_USER_META_APK 1
+#define BLAZE_USER_META_KIND_APK 1
 /**
  * [`blaze_user_meta_variant::elf`] is valid.
  */
+#define BLAZE_USER_META_KIND_ELF 2
+/**
+ * [`blaze_user_meta_variant::sym`] is valid.
+ */
+#define BLAZE_USER_META_KIND_SYM 3
+/**
+ * Deprecated; use `BLAZE_USER_META_KIND_UNKNOWN`.
+ */
+#define BLAZE_USER_META_KIND_BLAZE_USER_META_UNKNOWN 0
+/**
+ * Deprecated; use `BLAZE_USER_META_KIND_APK`.
+ */
+#define BLAZE_USER_META_KIND_BLAZE_USER_META_APK 1
+/**
+ * Deprecated; use `BLAZE_USER_META_KIND_ELF`.
+ */
 #define BLAZE_USER_META_KIND_BLAZE_USER_META_ELF 2
 
 /**
@@ -325,6 +354,188 @@ typedef struct blaze_user_meta_elf {
   uint8_t reserved[16];
 } blaze_user_meta_elf;
 
+/**
+ * Source code location information for a symbol or inlined function.
+ */
+typedef struct blaze_symbolize_code_info {
+  /**
+   * The directory in which the source file resides.
+   *
+   * This attribute is optional and may be NULL.
+   */
+  const char *dir;
+  /**
+   * The file that defines the symbol.
+   *
+   * This attribute is optional and may be NULL.
+   */
+  const char *file;
+  /**
+   * The line number on which the symbol is located in the source
+   * code.
+   */
+  uint32_t line;
+  /**
+   * The column number of the symbolized instruction in the source
+   * code.
+   */
+  uint16_t column;
+  /**
+   * Unused member available for future expansion.
+   */
+  uint8_t reserved[10];
+} blaze_symbolize_code_info;
+
+/**
+ * Data about an inlined function call.
+ */
+typedef struct blaze_symbolize_inlined_fn {
+  /**
+   * The symbol name of the inlined function.
+   */
+  const char *name;
+  /**
+   * Source code location information for the inlined function.
+   */
+  struct blaze_symbolize_code_info code_info;
+  /**
+   * Unused member available for future expansion.
+   */
+  uint8_t reserved[8];
+} blaze_symbolize_inlined_fn;
+
+/**
+ * The reason why symbolization failed.
+ *
+ * The reason is generally only meant as a hint. Reasons reported may
+ * change over time and, hence, should not be relied upon for the
+ * correctness of the application.
+ */
+typedef uint8_t blaze_symbolize_reason;
+/**
+ * Symbolization was successful.
+ */
+#define BLAZE_SYMBOLIZE_REASON_SUCCESS 0
+/**
+ * The absolute address was not found in the corresponding process'
+ * virtual memory map.
+ */
+#define BLAZE_SYMBOLIZE_REASON_UNMAPPED 1
+/**
+ * The file offset does not map to a valid piece of code/data.
+ */
+#define BLAZE_SYMBOLIZE_REASON_INVALID_FILE_OFFSET 2
+/**
+ * The `/proc/<pid>/maps` entry corresponding to the address does
+ * not have a component (file system path, object, ...) associated
+ * with it.
+ */
+#define BLAZE_SYMBOLIZE_REASON_MISSING_COMPONENT 3
+/**
+ * The symbolization source has no or no relevant symbols.
+ */
+#define BLAZE_SYMBOLIZE_REASON_MISSING_SYMS 4
+/**
+ * The address could not be found in the symbolization source.
+ */
+#define BLAZE_SYMBOLIZE_REASON_UNKNOWN_ADDR 5
+/**
+ * The address belonged to an entity that is currently unsupported.
+ */
+#define BLAZE_SYMBOLIZE_REASON_UNSUPPORTED 6
+
+/**
+ * The result of symbolization of an address.
+ *
+ * A `blaze_sym` is the information of a symbol found for an
+ * address.
+ */
+typedef struct blaze_sym {
+  /**
+   * The symbol name is where the given address should belong to.
+   *
+   * If an address could not be symbolized, this member will be NULL.
+   * Check the `reason` member for additional information pertaining
+   * the failure.
+   */
+  const char *name;
+  /**
+   * The path to or name of the module containing the symbol.
+   *
+   * Typically this would be the path to a executable or shared
+   * object. Depending on the symbol source this member may not be
+   * present or it could also just be a file name without path. In
+   * case of an ELF file contained inside an APK, this will be an
+   * Android style path of the form `<apk>!<elf-in-apk>`. E.g.,
+   * `/root/test.apk!/lib/libc.so`.
+   */
+  const char *module;
+  /**
+   * The address at which the symbol is located (i.e., its "start").
+   *
+   * This is the "normalized" address of the symbol, as present in
+   * the file (and reported by tools such as `readelf(1)`,
+   * `llvm-gsymutil`, or similar).
+   */
+  uint64_t addr;
+  /**
+   * The byte offset of the address that got symbolized from the
+   * start of the symbol (i.e., from `addr`).
+   *
+   * E.g., when symbolizing address 0x1337 of a function that starts at
+   * 0x1330, the offset will be set to 0x07 (and `addr` will be 0x1330). This
+   * member is especially useful in contexts when input addresses are not
+   * already normalized, such as when symbolizing an address in a process
+   * context (which may have been relocated and/or have layout randomizations
+   * applied).
+   */
+  size_t offset;
+  /**
+   * The size of the symbol.
+   *
+   * If the symbol's size is not available, this member will be `-1`.
+   * Note that some symbol sources may not distinguish between
+   * "unknown" size and `0`. In that case the size will be reported
+   * as `0` here as well.
+   */
+  ptrdiff_t size;
+  /**
+   * Source code location information for the symbol.
+   */
+  struct blaze_symbolize_code_info code_info;
+  /**
+   * The number of symbolized inlined function calls present.
+   */
+  size_t inlined_cnt;
+  /**
+   * An array of `inlined_cnt` symbolized inlined function calls.
+   */
+  const struct blaze_symbolize_inlined_fn *inlined;
+  /**
+   * On error (i.e., if `name` is NULL), a reason trying to explain
+   * why symbolization failed.
+   */
+  blaze_symbolize_reason reason;
+  /**
+   * Unused member available for future expansion.
+   */
+  uint8_t reserved[15];
+} blaze_sym;
+
+/**
+ * Readily symbolized information for an address.
+ */
+typedef struct blaze_user_meta_sym {
+  /**
+   * The symbol data.
+   */
+  const struct blaze_sym *sym;
+  /**
+   * Unused member available for future expansion.
+   */
+  uint8_t reserved[16];
+} blaze_user_meta_sym;
+
 /**
  * C compatible version of [`Unknown`].
  */
@@ -347,15 +558,19 @@ typedef struct blaze_user_meta_unknown {
  */
 typedef union blaze_user_meta_variant {
   /**
-   * Valid on [`blaze_user_meta_kind::BLAZE_USER_META_APK`].
+   * Valid on [`blaze_user_meta_kind::APK`].
    */
   struct blaze_user_meta_apk apk;
   /**
-   * Valid on [`blaze_user_meta_kind::BLAZE_USER_META_ELF`].
+   * Valid on [`blaze_user_meta_kind::ELF`].
    */
   struct blaze_user_meta_elf elf;
   /**
-   * Valid on [`blaze_user_meta_kind::BLAZE_USER_META_UNKNOWN`].
+   * Valid on [`blaze_user_meta_kind::SYM`].
+   */
+  struct blaze_user_meta_sym sym;
+  /**
+   * Valid on [`blaze_user_meta_kind::UNKNOWN`].
    */
   struct blaze_user_meta_unknown unknown;
 } blaze_user_meta_variant;
@@ -465,12 +680,10 @@ typedef struct blaze_normalize_opts {
   bool map_files;
   /**
    * Normalize addresses inside APKs to the contained ELF file and
-   * report a regular
-   * [`BLAZE_USER_META_ELF`][blaze_user_meta_kind::BLAZE_USER_META_ELF]
-   * meta data entry instead of an
-   * [`BLAZE_USER_META_APK`][blaze_user_meta_kind::BLAZE_USER_META_APK]
-   * one. As a result, the reported file offset will also be relative
-   * to the contained ELF file and not to the APK itself.
+   * report a regular [`blaze_user_meta_kind::ELF`] meta data entry
+   * instead of an [`blaze_user_meta_kind::APK`] one. As a result,
+   * the reported file offset will also be relative to the contained
+   * ELF file and not to the APK itself.
    */
   bool apk_to_elf;
   /**
@@ -480,46 +693,6 @@ typedef struct blaze_normalize_opts {
   uint8_t reserved[21];
 } blaze_normalize_opts;
 
-/**
- * The reason why symbolization failed.
- *
- * The reason is generally only meant as a hint. Reasons reported may
- * change over time and, hence, should not be relied upon for the
- * correctness of the application.
- */
-typedef uint8_t blaze_symbolize_reason;
-/**
- * Symbolization was successful.
- */
-#define BLAZE_SYMBOLIZE_REASON_SUCCESS 0
-/**
- * The absolute address was not found in the corresponding process'
- * virtual memory map.
- */
-#define BLAZE_SYMBOLIZE_REASON_UNMAPPED 1
-/**
- * The file offset does not map to a valid piece of code/data.
- */
-#define BLAZE_SYMBOLIZE_REASON_INVALID_FILE_OFFSET 2
-/**
- * The `/proc/<pid>/maps` entry corresponding to the address does
- * not have a component (file system path, object, ...) associated
- * with it.
- */
-#define BLAZE_SYMBOLIZE_REASON_MISSING_COMPONENT 3
-/**
- * The symbolization source has no or no relevant symbols.
- */
-#define BLAZE_SYMBOLIZE_REASON_MISSING_SYMS 4
-/**
- * The address could not be found in the symbolization source.
- */
-#define BLAZE_SYMBOLIZE_REASON_UNKNOWN_ADDR 5
-/**
- * The address belonged to an entity that is currently unsupported.
- */
-#define BLAZE_SYMBOLIZE_REASON_UNSUPPORTED 6
-
 /**
  * C ABI compatible version of [`blazesym::symbolize::Symbolizer`].
  *
@@ -553,7 +726,7 @@ typedef struct blaze_symbolizer_opts {
    */
   const char *const *debug_dirs;
   /**
-   * The number of array elements in `debug_dirs`.
+   * The number of array elements in [`debug_dirs`][Self::debug_dirs].
    */
   size_t debug_dirs_len;
   /**
@@ -565,11 +738,17 @@ typedef struct blaze_symbolizer_opts {
   /**
    * Whether to attempt to gather source code location information.
    *
-   * This setting implies `debug_syms` (and forces it to `true`).
+   * This option only has an effect if `debug_syms` of the particular
+   * symbol source is set to `true`. Furthermore, it is a necessary
+   * prerequisite for retrieving inlined function information (see
+   * [`inlined_fns`][Self::inlined_fns]).
    */
   bool code_info;
   /**
    * Whether to report inlined functions as part of symbolization.
+   *
+   * This option only has an effect if [`code_info`][Self::code_info]
+   * is `true`.
    */
   bool inlined_fns;
   /**
@@ -652,134 +831,6 @@ typedef struct blaze_cache_src_process {
   uint8_t reserved[19];
 } blaze_cache_src_process;
 
-/**
- * Source code location information for a symbol or inlined function.
- */
-typedef struct blaze_symbolize_code_info {
-  /**
-   * The directory in which the source file resides.
-   *
-   * This attribute is optional and may be NULL.
-   */
-  const char *dir;
-  /**
-   * The file that defines the symbol.
-   *
-   * This attribute is optional and may be NULL.
-   */
-  const char *file;
-  /**
-   * The line number on which the symbol is located in the source
-   * code.
-   */
-  uint32_t line;
-  /**
-   * The column number of the symbolized instruction in the source
-   * code.
-   */
-  uint16_t column;
-  /**
-   * Unused member available for future expansion.
-   */
-  uint8_t reserved[10];
-} blaze_symbolize_code_info;
-
-/**
- * Data about an inlined function call.
- */
-typedef struct blaze_symbolize_inlined_fn {
-  /**
-   * The symbol name of the inlined function.
-   */
-  const char *name;
-  /**
-   * Source code location information for the inlined function.
-   */
-  struct blaze_symbolize_code_info code_info;
-  /**
-   * Unused member available for future expansion.
-   */
-  uint8_t reserved[8];
-} blaze_symbolize_inlined_fn;
-
-/**
- * The result of symbolization of an address.
- *
- * A `blaze_sym` is the information of a symbol found for an
- * address.
- */
-typedef struct blaze_sym {
-  /**
-   * The symbol name is where the given address should belong to.
-   *
-   * If an address could not be symbolized, this member will be NULL.
-   * Check the `reason` member for additional information pertaining
-   * the failure.
-   */
-  const char *name;
-  /**
-   * The path to or name of the module containing the symbol.
-   *
-   * Typically this would be the path to a executable or shared
-   * object. Depending on the symbol source this member may not be
-   * present or it could also just be a file name without path. In
-   * case of an ELF file contained inside an APK, this will be an
-   * Android style path of the form `<apk>!<elf-in-apk>`. E.g.,
-   * `/root/test.apk!/lib/libc.so`.
-   */
-  const char *module;
-  /**
-   * The address at which the symbol is located (i.e., its "start").
-   *
-   * This is the "normalized" address of the symbol, as present in
-   * the file (and reported by tools such as `readelf(1)`,
-   * `llvm-gsymutil`, or similar).
-   */
-  uint64_t addr;
-  /**
-   * The byte offset of the address that got symbolized from the
-   * start of the symbol (i.e., from `addr`).
-   *
-   * E.g., when symbolizing address 0x1337 of a function that starts at
-   * 0x1330, the offset will be set to 0x07 (and `addr` will be 0x1330). This
-   * member is especially useful in contexts when input addresses are not
-   * already normalized, such as when symbolizing an address in a process
-   * context (which may have been relocated and/or have layout randomizations
-   * applied).
-   */
-  size_t offset;
-  /**
-   * The size of the symbol.
-   *
-   * If the symbol's size is not available, this member will be `-1`.
-   * Note that some symbol sources may not distinguish between
-   * "unknown" size and `0`. In that case the size will be reported
-   * as `0` here as well.
-   */
-  ptrdiff_t size;
-  /**
-   * Source code location information for the symbol.
-   */
-  struct blaze_symbolize_code_info code_info;
-  /**
-   * The number of symbolized inlined function calls present.
-   */
-  size_t inlined_cnt;
-  /**
-   * An array of `inlined_cnt` symbolized inlined function calls.
-   */
-  const struct blaze_symbolize_inlined_fn *inlined;
-  /**
-   * On error (i.e., if `name` is NULL), a reason trying to explain
-   * why symbolization failed.
-   */
-  blaze_symbolize_reason reason;
-  /**
-   * Unused member available for future expansion.
-   */
-  uint8_t reserved[15];
-} blaze_sym;
-
 /**
  * `blaze_syms` is the result of symbolization of a list of addresses.
  *
@@ -839,11 +890,16 @@ typedef struct blaze_symbolize_src_process {
    * the need for requiring the `SYS_ADMIN` capability is eliminated.
    */
   bool no_map_files;
+  /**
+   * Whether or not to symbolize addresses in a vDSO (virtual dynamic
+   * shared object).
+   */
+  bool no_vdso;
   /**
    * Unused member available for future expansion. Must be initialized
    * to zero.
    */
-  uint8_t reserved[17];
+  uint8_t reserved[16];
 } blaze_symbolize_src_process;
 
 /**
@@ -1174,7 +1230,7 @@ void blaze_normalizer_free(blaze_normalizer *normalizer);
 /**
  * Retrieve a textual representation of the reason of a normalization failure.
  */
-const char *blaze_normalize_reason_str(blaze_normalize_reason err);
+const char *blaze_normalize_reason_str(blaze_normalize_reason reason);
 
 /**
  * Normalize a list of user space addresses.
@@ -1242,7 +1298,7 @@ void blaze_user_output_free(struct blaze_normalized_user_output *output);
  * Retrieve a textual representation of the reason of a symbolization
  * failure.
  */
-const char *blaze_symbolize_reason_str(blaze_symbolize_reason err);
+const char *blaze_symbolize_reason_str(blaze_symbolize_reason reason);
 
 /**
  * Create an instance of a symbolizer.
