@@ -20,7 +20,6 @@ use blazesym::symbolize::source::Kernel;
 use blazesym::symbolize::source::Process;
 use blazesym::symbolize::source::Source;
 use blazesym::symbolize::CodeInfo;
-use blazesym::symbolize::InlinedFn;
 use blazesym::symbolize::Input;
 use blazesym::symbolize::Reason;
 use blazesym::symbolize::Sym;
@@ -35,6 +34,7 @@ use crate::blaze_err_last;
 use crate::set_last_err;
 use crate::util::slice_from_aligned_user_array;
 use crate::util::slice_from_user_array;
+use crate::util::DynSize as _;
 
 
 /// Configuration for caching of ELF symbolization data.
@@ -313,9 +313,12 @@ pub struct blaze_symbolize_src_process {
     /// using symbolic paths (i.e., with `no_map_files` being `true`)
     /// the need for requiring the `SYS_ADMIN` capability is eliminated.
     pub no_map_files: bool,
+    /// Whether or not to symbolize addresses in a vDSO (virtual dynamic
+    /// shared object).
+    pub no_vdso: bool,
     /// Unused member available for future expansion. Must be initialized
     /// to zero.
-    pub reserved: [u8; 17],
+    pub reserved: [u8; 16],
 }
 
 impl Default for blaze_symbolize_src_process {
@@ -326,7 +329,8 @@ impl Default for blaze_symbolize_src_process {
             debug_syms: false,
             perf_map: false,
             no_map_files: false,
-            reserved: [0; 17],
+            no_vdso: false,
+            reserved: [0; 16],
         }
     }
 }
@@ -339,6 +343,7 @@ impl From<blaze_symbolize_src_process> for Process {
             debug_syms,
             perf_map,
             no_map_files,
+            no_vdso,
             reserved: _,
         } = process;
         Self {
@@ -346,6 +351,7 @@ impl From<blaze_symbolize_src_process> for Process {
             debug_syms,
             perf_map,
             map_files: !no_map_files,
+            vdso: !no_vdso,
             _non_exhaustive: (),
         }
     }
@@ -454,6 +460,8 @@ pub type blaze_symbolizer = Symbolizer;
 #[derive(Copy, Clone, Debug, Eq, PartialEq)]
 pub struct blaze_symbolize_reason(u8);
 
+// Please adjust `blaze_normalize_reason` as well when adding a new
+// variant.
 impl blaze_symbolize_reason {
     /// Symbolization was successful.
     pub const SUCCESS: blaze_symbolize_reason = blaze_symbolize_reason(0);
@@ -493,8 +501,8 @@ impl From<Reason> for blaze_symbolize_reason {
 /// Retrieve a textual representation of the reason of a symbolization
 /// failure.
 #[no_mangle]
-pub extern "C" fn blaze_symbolize_reason_str(err: blaze_symbolize_reason) -> *const c_char {
-    match err {
+pub extern "C" fn blaze_symbolize_reason_str(reason: blaze_symbolize_reason) -> *const c_char {
+    match reason {
         blaze_symbolize_reason::SUCCESS => b"success\0".as_ptr().cast(),
         blaze_symbolize_reason::UNMAPPED => Reason::Unmapped.as_bytes().as_ptr().cast(),
         blaze_symbolize_reason::INVALID_FILE_OFFSET => {
@@ -653,7 +661,7 @@ pub struct blaze_symbolizer_opts {
     /// Note that the directory containing a symbolization source is always an
     /// implicit candidate target directory of the highest precedence.
     pub debug_dirs: *const *const c_char,
-    /// The number of array elements in `debug_dirs`.
+    /// The number of array elements in [`debug_dirs`][Self::debug_dirs].
     pub debug_dirs_len: usize,
     /// Whether or not to automatically reload file system based
     /// symbolization sources that were updated since the last
@@ -661,9 +669,15 @@ pub struct blaze_symbolizer_opts {
     pub auto_reload: bool,
     /// Whether to attempt to gather source code location information.
     ///
-    /// This setting implies `debug_syms` (and forces it to `true`).
+    /// This option only has an effect if `debug_syms` of the particular
+    /// symbol source is set to `true`. Furthermore, it is a necessary
+    /// prerequisite for retrieving inlined function information (see
+    /// [`inlined_fns`][Self::inlined_fns]).
     pub code_info: bool,
     /// Whether to report inlined functions as part of symbolization.
+    ///
+    /// This option only has an effect if [`code_info`][Self::code_info]
+    /// is `true`.
     pub inlined_fns: bool,
     /// Whether or not to transparently demangle symbols.
     ///
@@ -869,49 +883,31 @@ pub unsafe extern "C" fn blaze_symbolize_cache_process(
     let () = set_last_err(err);
 }
 
-fn code_info_strtab_size(code_info: &Option<CodeInfo>) -> usize {
-    code_info
-        .as_ref()
-        .and_then(|info| info.dir.as_ref().map(|d| d.as_os_str().len() + 1))
-        .unwrap_or(0)
-        + code_info
-            .as_ref()
-            .map(|info| info.file.len() + 1)
-            .unwrap_or(0)
-}
+fn make_cstr(src: &OsStr, cstr_last: &mut *mut c_char) -> *mut c_char {
+    let cstr = *cstr_last;
+    unsafe { ptr::copy_nonoverlapping(src.as_bytes().as_ptr(), cstr as *mut u8, src.len()) };
+    unsafe { *cstr.add(src.len()) = 0 };
+    *cstr_last = unsafe { cstr_last.add(src.len() + 1) };
 
-fn inlined_fn_strtab_size(inlined_fn: &InlinedFn) -> usize {
-    inlined_fn.name.len() + 1 + code_info_strtab_size(&inlined_fn.code_info)
-}
-
-fn sym_strtab_size(sym: &Sym) -> usize {
-    sym.name.len()
-        + 1
-        + sym
-            .module
-            .as_deref()
-            .map(|module| module.len() + 1)
-            .unwrap_or(0)
-        + code_info_strtab_size(&sym.code_info)
-        + sym
-            .inlined
-            .iter()
-            .map(inlined_fn_strtab_size)
-            .sum::<usize>()
+    cstr
 }
 
 fn convert_code_info(
     code_info_in: &Option<CodeInfo>,
     code_info_out: &mut blaze_symbolize_code_info,
-    mut make_cstr: impl FnMut(&OsStr) -> *mut c_char,
+    cstr_last: &mut *mut c_char,
 ) {
     code_info_out.dir = code_info_in
         .as_ref()
-        .and_then(|info| info.dir.as_ref().map(|d| make_cstr(d.as_os_str())))
+        .and_then(|info| {
+            info.dir
+                .as_ref()
+                .map(|d| make_cstr(d.as_os_str(), cstr_last))
+        })
         .unwrap_or_else(ptr::null_mut);
     code_info_out.file = code_info_in
         .as_ref()
-        .map(|info| make_cstr(&info.file))
+        .map(|info| make_cstr(&info.file, cstr_last))
         .unwrap_or_else(ptr::null_mut);
     code_info_out.line = code_info_in
         .as_ref()
@@ -923,6 +919,46 @@ fn convert_code_info(
         .unwrap_or(0);
 }
 
+
+/// Convert a [`Sym`] into the `blaze_sym` C correspondent.
+pub(crate) fn convert_sym(
+    sym: &Sym,
+    sym_ref: &mut blaze_sym,
+    inlined_last: &mut *mut blaze_symbolize_inlined_fn,
+    cstr_last: &mut *mut c_char,
+) {
+    let name_ptr = make_cstr(OsStr::new(sym.name.as_ref()), cstr_last);
+    let module_ptr = sym
+        .module
+        .as_deref()
+        .map(|module| make_cstr(module, cstr_last))
+        .unwrap_or(ptr::null_mut());
+
+    sym_ref.name = name_ptr;
+    sym_ref.module = module_ptr;
+    sym_ref.addr = sym.addr;
+    sym_ref.offset = sym.offset;
+    sym_ref.size = sym
+        .size
+        .map(|size| isize::try_from(size).unwrap_or(isize::MAX))
+        .unwrap_or(-1);
+    convert_code_info(&sym.code_info, &mut sym_ref.code_info, cstr_last);
+    sym_ref.inlined_cnt = sym.inlined.len();
+    sym_ref.inlined = *inlined_last;
+    sym_ref.reason = blaze_symbolize_reason::SUCCESS;
+
+    for inlined in sym.inlined.iter() {
+        let inlined_ref = unsafe { &mut **inlined_last };
+
+        let name_ptr = make_cstr(OsStr::new(inlined.name.as_ref()), cstr_last);
+        inlined_ref.name = name_ptr;
+        convert_code_info(&inlined.code_info, &mut inlined_ref.code_info, cstr_last);
+
+        *inlined_last = unsafe { inlined_last.add(1) };
+    }
+}
+
+
 /// Convert [`Sym`] objects to [`blaze_syms`] ones.
 ///
 /// The returned pointer should be released using [`blaze_syms_free`] once
@@ -931,7 +967,7 @@ fn convert_symbolizedresults_to_c(results: Vec<Symbolized>) -> *const blaze_syms
     // Allocate a buffer to contain a blaze_syms, all
     // blaze_sym, and C strings of symbol and path.
     let (strtab_size, inlined_fn_cnt) = results.iter().fold((0, 0), |acc, sym| match sym {
-        Symbolized::Sym(sym) => (acc.0 + sym_strtab_size(sym), acc.1 + sym.inlined.len()),
+        Symbolized::Sym(sym) => (acc.0 + sym.c_str_size(), acc.1 + sym.inlined.len()),
         Symbolized::Unknown(..) => acc,
     });
 
@@ -940,38 +976,29 @@ fn convert_symbolizedresults_to_c(results: Vec<Symbolized>) -> *const blaze_syms
         + mem::size_of::<blaze_syms>()
         + mem::size_of::<blaze_sym>() * results.len()
         + mem::size_of::<blaze_symbolize_inlined_fn>() * inlined_fn_cnt;
-    let raw_buf_with_sz = unsafe { alloc(Layout::from_size_align(buf_size, 8).unwrap()) };
-    if raw_buf_with_sz.is_null() {
+    let buf = unsafe { alloc(Layout::from_size_align(buf_size, 8).unwrap()) };
+    if buf.is_null() {
         return ptr::null()
     }
 
-    // prepend an u64 to keep the size of the buffer.
-    unsafe { *(raw_buf_with_sz as *mut u64) = buf_size as u64 };
+    // Prepend a `u64` to store the size of the buffer.
+    unsafe { *buf.cast::<u64>() = buf_size as u64 };
 
-    let raw_buf = unsafe { raw_buf_with_sz.add(mem::size_of::<u64>()) };
+    let syms_buf = unsafe { buf.add(mem::size_of::<u64>()) };
 
-    let syms_ptr = raw_buf as *mut blaze_syms;
-    let mut syms_last = unsafe { (*syms_ptr).syms.as_mut_slice().as_mut_ptr() };
+    let syms_ptr = syms_buf.cast::<blaze_syms>();
+    let mut syms_last = unsafe { (*syms_ptr).syms.as_mut_ptr() };
     let mut inlined_last = unsafe {
-        raw_buf.add(mem::size_of::<blaze_syms>() + mem::size_of::<blaze_sym>() * results.len())
+        syms_buf.add(mem::size_of::<blaze_syms>() + mem::size_of::<blaze_sym>() * results.len())
     } as *mut blaze_symbolize_inlined_fn;
     let mut cstr_last = unsafe {
-        raw_buf.add(
+        syms_buf.add(
             mem::size_of::<blaze_syms>()
                 + mem::size_of::<blaze_sym>() * results.len()
                 + mem::size_of::<blaze_symbolize_inlined_fn>() * inlined_fn_cnt,
         )
     } as *mut c_char;
 
-    let mut make_cstr = |src: &OsStr| {
-        let cstr = cstr_last;
-        unsafe { ptr::copy_nonoverlapping(src.as_bytes().as_ptr(), cstr as *mut u8, src.len()) };
-        unsafe { *cstr.add(src.len()) = 0 };
-        cstr_last = unsafe { cstr_last.add(src.len() + 1) };
-
-        cstr
-    };
-
     unsafe { (*syms_ptr).cnt = results.len() };
 
     // Convert all `Sym`s to `blazesym_sym`s.
@@ -979,39 +1006,7 @@ fn convert_symbolizedresults_to_c(results: Vec<Symbolized>) -> *const blaze_syms
         match sym {
             Symbolized::Sym(sym) => {
                 let sym_ref = unsafe { &mut *syms_last };
-                let name_ptr = make_cstr(OsStr::new(sym.name.as_ref()));
-                let module_ptr = sym
-                    .module
-                    .as_deref()
-                    .map(&mut make_cstr)
-                    .unwrap_or(ptr::null_mut());
-
-                sym_ref.name = name_ptr;
-                sym_ref.module = module_ptr;
-                sym_ref.addr = sym.addr;
-                sym_ref.offset = sym.offset;
-                sym_ref.size = sym
-                    .size
-                    .map(|size| isize::try_from(size).unwrap_or(isize::MAX))
-                    .unwrap_or(-1);
-                convert_code_info(&sym.code_info, &mut sym_ref.code_info, &mut make_cstr);
-                sym_ref.inlined_cnt = sym.inlined.len();
-                sym_ref.inlined = inlined_last;
-                sym_ref.reason = blaze_symbolize_reason::SUCCESS;
-
-                for inlined in sym.inlined.iter() {
-                    let inlined_ref = unsafe { &mut *inlined_last };
-
-                    let name_ptr = make_cstr(OsStr::new(inlined.name.as_ref()));
-                    inlined_ref.name = name_ptr;
-                    convert_code_info(
-                        &inlined.code_info,
-                        &mut inlined_ref.code_info,
-                        &mut make_cstr,
-                    );
-
-                    inlined_last = unsafe { inlined_last.add(1) };
-                }
+                let () = convert_sym(&sym, sym_ref, &mut inlined_last, &mut cstr_last);
             }
             Symbolized::Unknown(reason) => {
                 // Unknown symbols/addresses are just represented with all
@@ -1306,9 +1301,10 @@ pub unsafe extern "C" fn blaze_syms_free(syms: *const blaze_syms) {
         return
     }
 
-    let raw_buf_with_sz = unsafe { (syms as *mut u8).offset(-(mem::size_of::<u64>() as isize)) };
-    let sz = unsafe { *(raw_buf_with_sz as *mut u64) } as usize;
-    unsafe { dealloc(raw_buf_with_sz, Layout::from_size_align(sz, 8).unwrap()) };
+    // Retrieve back the buffer with the `u64` size header.
+    let buf = unsafe { syms.byte_sub(mem::size_of::<u64>()).cast::<u8>().cast_mut() };
+    let size = unsafe { *buf.cast::<u64>() } as usize;
+    unsafe { dealloc(buf, Layout::from_size_align(size, 8).unwrap()) };
 }
 
 
@@ -1328,6 +1324,7 @@ mod tests {
 
     use blazesym::inspect;
     use blazesym::normalize;
+    use blazesym::symbolize::InlinedFn;
     use blazesym::symbolize::Reason;
     use blazesym::Pid;
 
@@ -1387,7 +1384,7 @@ mod tests {
         };
         assert_eq!(
             format!("{process:?}"),
-            "blaze_symbolize_src_process { type_size: 16, pid: 1337, debug_syms: true, perf_map: false, no_map_files: false, reserved: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0] }"
+            "blaze_symbolize_src_process { type_size: 16, pid: 1337, debug_syms: true, perf_map: false, no_map_files: false, no_vdso: false, reserved: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0] }"
         );
 
         let gsym_data = blaze_symbolize_src_gsym_data {
@@ -1963,8 +1960,7 @@ mod tests {
             let name = unsafe { CStr::from_ptr(sym.name) };
             assert!(
                 name.to_str().unwrap().contains("test13test_function"),
-                "{:?}",
-                name
+                "{name:?}"
             );
 
             if sym.inlined_cnt == 0 {
@@ -1977,8 +1973,7 @@ mod tests {
             let name = unsafe { CStr::from_ptr((*sym.inlined).name) };
             assert!(
                 name.to_str().unwrap().contains("test12inlined_call"),
-                "{:?}",
-                name
+                "{name:?}"
             );
 
             let () = unsafe { blaze_syms_free(result) };
