@@ -16,6 +16,8 @@ use std::rc::Rc;
 use gimli::AbbreviationsCacheStrategy;
 use gimli::Dwarf;
 
+use crate::dwarf::reader::Endianess;
+use crate::dwarf::reader::R;
 use crate::elf::ElfParser;
 #[cfg(test)]
 use crate::elf::DEFAULT_DEBUG_DIRS;
@@ -24,6 +26,7 @@ use crate::inspect::FindAddrOpts;
 use crate::inspect::ForEachFn;
 use crate::inspect::Inspect;
 use crate::inspect::SymInfo;
+use crate::log;
 use crate::log::debug;
 use crate::log::warn;
 use crate::symbolize::CodeInfo;
@@ -36,6 +39,7 @@ use crate::symbolize::Symbolize;
 use crate::Addr;
 use crate::Error;
 use crate::ErrorExt;
+use crate::ErrorKind;
 use crate::Mmap;
 use crate::Result;
 use crate::SymType;
@@ -78,7 +82,9 @@ impl From<Option<gimli::DwLang>> for SrcLang {
                 gimli::DW_LANG_C_plus_plus
                 | gimli::DW_LANG_C_plus_plus_03
                 | gimli::DW_LANG_C_plus_plus_11
-                | gimli::DW_LANG_C_plus_plus_14,
+                | gimli::DW_LANG_C_plus_plus_14
+                | gimli::DW_LANG_C_plus_plus_17
+                | gimli::DW_LANG_C_plus_plus_20,
             ) => SrcLang::Cpp,
             _ => SrcLang::Unknown,
         }
@@ -112,6 +118,9 @@ fn find_debug_file(file: &OsStr, linker: Option<&Path>, debug_dirs: &[PathBuf])
 }
 
 
+// TODO: We really should have a cache of debug link targets as well, to
+//       avoid potentially re-parsing the same files over and over
+//       again.
 fn try_deref_debug_link(
     parser: &ElfParser,
     debug_dirs: &[PathBuf],
@@ -148,7 +157,29 @@ fn try_deref_debug_link(
 }
 
 
-/// DwarfResolver provides abilities to query DWARF information of binaries.
+/// Try to find a DWARF package (`.dwp`) "belonging" to the file
+/// referenced by the given [`ElfParser`].
+fn try_find_dwp(parser: &ElfParser) -> Result<Option<Rc<ElfParser>>> {
+    if let Some(path) = parser.module() {
+        let mut dwp_path = path.to_os_string();
+        let () = dwp_path.push(".dwp");
+        let dwp_path = PathBuf::from(dwp_path);
+
+        match ElfParser::open(&dwp_path) {
+            Ok(parser) => {
+                log::debug!("using DWARF package `{}`", dwp_path.display());
+                Ok(Some(Rc::new(parser)))
+            }
+            Err(err) if err.kind() == ErrorKind::NotFound => Ok(None),
+            Err(err) => Err(err),
+        }
+    } else {
+        Ok(None)
+    }
+}
+
+
+/// `DwarfResolver` provides abilities to query DWARF information of binaries.
 pub(crate) struct DwarfResolver {
     /// The lazily parsed compilation units of the DWARF file.
     // SAFETY: We must not hand out references with a 'static lifetime to
@@ -161,6 +192,9 @@ pub(crate) struct DwarfResolver {
     /// If the source file contains a valid debug link, this parser
     /// represents it.
     linkee_parser: Option<Rc<ElfParser>>,
+    /// If there exist an associated DWARF Package (*.dwp), this parser
+    /// represents it.
+    _dwp_parser: Option<Rc<ElfParser>>,
 }
 
 impl DwarfResolver {
@@ -169,21 +203,19 @@ impl DwarfResolver {
         &self.parser
     }
 
-    pub(crate) fn from_parser(
-        parser: Rc<ElfParser>,
-        debug_dirs: &[PathBuf],
-    ) -> Result<Self, Error> {
+    pub(crate) fn from_parser(parser: Rc<ElfParser>, debug_dirs: &[PathBuf]) -> Result<Self> {
         let linkee_parser = try_deref_debug_link(&parser, debug_dirs)?;
+        let dwp_parser = try_find_dwp(&parser)?;
 
         // SAFETY: We own the `ElfParser` and make sure that it stays
         //         around while the `Units` object uses it. As such, it
         //         is fine to conjure a 'static lifetime here.
-        let static_parser = unsafe {
+        let static_linkee_parser = unsafe {
             mem::transmute::<&ElfParser, &'static ElfParser>(
                 linkee_parser.as_ref().unwrap_or(&parser).deref(),
             )
         };
-        let mut load_section = |section| reader::load_section(static_parser, section);
+        let mut load_section = |section| reader::load_section(static_linkee_parser, section);
         let mut dwarf = Dwarf::load(&mut load_section)?;
         // Cache abbreviations (which will cause them to be
         // automatically reused across compilation units), which can
@@ -192,11 +224,29 @@ impl DwarfResolver {
         // much effort the linker spent on optimizing it.
         let () = dwarf.populate_abbreviations_cache(AbbreviationsCacheStrategy::Duplicates);
 
-        let units = Units::parse(dwarf)?;
+        let dwp = dwp_parser
+            .as_deref()
+            .map(|dwp_parser| {
+                let empty = R::new(&[], Endianess::default());
+                // SAFETY: We own the `ElfParser` and make sure that it
+                //         stays around while the `Units` object uses
+                //         it. As such, it is fine to conjure a 'static
+                //         lifetime here.
+                let static_dwp_parser =
+                    unsafe { mem::transmute::<&ElfParser, &'static ElfParser>(dwp_parser) };
+                let load_dwo_section =
+                    |section| reader::load_dwo_section(static_dwp_parser, section);
+                let dwp = gimli::DwarfPackage::load(load_dwo_section, empty)?;
+                Result::<_, Error>::Ok(dwp)
+            })
+            .transpose()?;
+
+        let units = Units::parse(dwarf, dwp)?;
         let slf = Self {
             units,
             parser,
             linkee_parser,
+            _dwp_parser: dwp_parser,
         };
         Ok(slf)
     }
@@ -230,7 +280,7 @@ impl DwarfResolver {
         let addr = function
             .range
             .as_ref()
-            .map(|range| range.begin as Addr)
+            .map(|range| range.begin)
             .unwrap_or(0);
         let size = function
             .range
@@ -244,10 +294,11 @@ impl DwarfResolver {
             size: Some(size),
             sym_type: SymType::Function,
             file_offset: offset_in_file
-                .then(|| self.parser.find_file_offset(addr))
+                .then(|| self.parser.find_file_offset(addr, size))
                 .transpose()?
                 .flatten(),
             module: self.parser.module().map(Cow::Borrowed),
+            _non_exhaustive: (),
         };
         Ok(Some(info))
     }
@@ -520,6 +571,24 @@ mod tests {
         );
     }
 
+    /// Check that we can discover DWARF packages as expected.
+    #[test]
+    fn dwp_discovery() {
+        let path = Path::new(&env!("CARGO_MANIFEST_DIR"))
+            .join("data")
+            .join("test-rs-split-dwarf.bin");
+        let parser = ElfParser::open(&path).unwrap();
+        let dwp_parser = try_find_dwp(&parser).unwrap();
+        assert!(dwp_parser.is_some());
+
+        let path = Path::new(&env!("CARGO_MANIFEST_DIR"))
+            .join("data")
+            .join("test-rs.bin");
+        let parser = ElfParser::open(&path).unwrap();
+        let dwp_parser = try_find_dwp(&parser).unwrap();
+        assert!(dwp_parser.is_none());
+    }
+
     /// Check that we can find the source code location of an address.
     #[test]
     fn source_location_finding() {
