@@ -1,11 +1,12 @@
 use std::borrow::Cow;
+use std::cell::OnceCell;
 use std::ffi::OsStr;
 use std::ffi::OsString;
 use std::fmt::Debug;
 use std::fmt::Formatter;
 use std::fmt::Result as FmtResult;
 use std::fs::File;
-use std::io::Read as _;
+use std::io;
 use std::io::Seek as _;
 use std::io::SeekFrom;
 use std::mem;
@@ -20,13 +21,14 @@ use crate::inspect::FindAddrOpts;
 use crate::inspect::ForEachFn;
 use crate::inspect::SymInfo;
 use crate::mmap::Mmap;
-use crate::once::OnceCell;
 use crate::pathlike::PathLike;
 use crate::symbolize::FindSymOpts;
 use crate::symbolize::Reason;
 use crate::symbolize::ResolvedSym;
 use crate::symbolize::SrcLang;
+use crate::symbolize::Symbolize;
 use crate::util::find_match_or_lower_bound_by_key;
+use crate::util::OnceCellExt as _;
 use crate::util::Pod;
 use crate::util::ReadRaw as _;
 use crate::Addr;
@@ -60,12 +62,63 @@ use super::types::ELFCOMPRESS_ZSTD;
 use super::types::PN_XNUM;
 use super::types::PT_LOAD;
 use super::types::SHF_COMPRESSED;
+use super::types::SHN_ABS;
 use super::types::SHN_LORESERVE;
 use super::types::SHN_UNDEF;
 use super::types::SHN_XINDEX;
 use super::types::SHT_NOBITS;
 
 
+pub(crate) type StaticMem = &'static [u8];
+
+
+fn read_pod<T, R>(reader: &mut R) -> Result<T, io::Error>
+where
+    T: Pod,
+    R: io::Read,
+{
+    let mut value = MaybeUninit::<T>::uninit();
+    let ptr = value.as_mut_ptr().cast::<u8>();
+    let len = mem::size_of::<T>();
+    // Make sure to zero out everything, including potential padding
+    // bytes. `std::io::Read` requires fully initialized memory, for
+    // better or worse.
+    // SAFETY: `T` is a `Pod` and hence valid for any bit pattern,
+    //         including all zeroes.
+    let () = unsafe { ptr.write_bytes(0, len) };
+    // SAFETY: `value` is a buffer of `len` bytes.
+    let slice = unsafe { slice::from_raw_parts_mut(ptr, len) };
+    let () = reader.read_exact(slice)?;
+    // SAFETY: `T` is a `Pod` and hence valid for any bit pattern,
+    //         including all zeroes.
+    Ok(unsafe { value.assume_init() })
+}
+
+
+fn read_pod_vec<T, R>(reader: &mut R, count: usize) -> Result<Vec<T>, io::Error>
+where
+    T: Pod,
+    R: io::Read,
+{
+    let mut vec = Vec::<T>::with_capacity(count);
+    let ptr = vec.as_mut_ptr().cast::<u8>();
+    let len = count * mem::size_of::<T>();
+    // Make sure to zero out everything, including potential padding
+    // bytes. `std::io::Read` requires fully initialized memory, for
+    // better or worse.
+    // SAFETY: `T` is a `Pod` and hence valid for any bit pattern,
+    //         including all zeroes.
+    let () = unsafe { ptr.write_bytes(0, len) };
+    // SAFETY: `vec` is a buffer of `len` bytes.
+    let slice = unsafe { slice::from_raw_parts_mut(ptr, len) };
+    let () = reader.read_exact(slice)?;
+    // SAFETY: The `Vec` is guaranteed to have capacity for `count`
+    //         objects and we made sure to initialize all of them.
+    let () = unsafe { vec.set_len(count) };
+    Ok(vec)
+}
+
+
 fn symbol_name<'elf>(strtab: &'elf [u8], sym: &Elf64_Sym) -> Result<&'elf str> {
     let name = strtab
         .get(sym.st_name as usize..)
@@ -90,7 +143,7 @@ fn find_sym<'elf>(
     let idx = find_match_or_lower_bound_by_key(by_addr_idx, addr, |&idx| {
         // SANITY: The index originates in our code and is known to be
         //         in bounds.
-        syms.get(idx).unwrap().value() as Addr
+        syms.get(idx).unwrap().value()
     });
 
     match idx {
@@ -100,7 +153,7 @@ fn find_sym<'elf>(
                 // SANITY: The index originates in our code and is known
                 //         to be in bounds.
                 let sym = syms.get(*idx).unwrap().to_64bit();
-                if sym.st_value as Addr > addr {
+                if sym.st_value > addr {
                     // Once we are seeing start addresses past the provided
                     // address, we can no longer be dealing with a match and
                     // stop the search.
@@ -117,7 +170,7 @@ fn find_sym<'elf>(
                     let sym = ResolvedSym {
                         name: symbol_name(strtab, &sym)?,
                         module,
-                        addr: sym.st_value as Addr,
+                        addr: sym.st_value,
                         size: if sym.st_size == 0 {
                             None
                         } else {
@@ -152,6 +205,10 @@ fn file_offset(shdrs: &ElfN_Shdrs<'_>, sym: &Elf64_Sym) -> Result<Option<u64>> {
         return Ok(None)
     }
 
+    if sym.st_shndx == SHN_ABS {
+        return Ok(None)
+    }
+
     let shdr = shdrs
         .get(usize::from(sym.st_shndx))
         .ok_or_invalid_input(|| {
@@ -161,11 +218,23 @@ fn file_offset(shdrs: &ElfN_Shdrs<'_>, sym: &Elf64_Sym) -> Result<Option<u64>> {
             )
         })?;
 
+    // If the section doesn't have in-file data, there can't be a proper
+    // file offset (elf(5) refers to a "conceptual placement in the
+    // file", but that concept has no meaning for us).
+    if shdr.type_() == SHT_NOBITS {
+        return Ok(None)
+    }
+
     let offset = sym
         .st_value
         .wrapping_sub(shdr.addr())
         .wrapping_add(shdr.offset());
-    Ok(Some(offset))
+    let limit = shdr.offset() + shdr.size();
+    if offset < limit || (offset == limit && sym.st_size == 0) {
+        Ok(Some(offset))
+    } else {
+        Ok(None)
+    }
 }
 
 
@@ -340,7 +409,7 @@ impl<'elf> SymbolTableCache<'elf> {
     {
         let str2sym = self
             .str2sym
-            .get_or_try_init(|| {
+            .get_or_try_init_(|| {
                 let str2sym = self.create_str2sym(filter)?;
                 Result::<_, Error>::Ok(str2sym)
             })?
@@ -400,9 +469,9 @@ where
     /// `idx`.
     ///
     /// # Notes
-    /// This method returns potentially compressed data, but adheres is
-    /// able to do so with 'elf lifetime. To transparently decompress,
-    /// use [`Cache::section_data`] instead.
+    /// This method returns potentially compressed data, but is able to
+    /// do so with the `'elf` lifetime. To transparently decompress, use
+    /// [`Cache::section_data`] instead.
     fn section_data_raw(&self, idx: usize) -> Result<Cow<'elf, [u8]>> {
         let shdr = self.section_hdr(idx)?;
         if shdr.type_() != SHT_NOBITS {
@@ -422,7 +491,7 @@ where
     fn section_data(&self, idx: usize) -> Result<&[u8]> {
         let shdr = self.section_hdr(idx)?;
         if shdr.type_() != SHT_NOBITS {
-            let datas = self.section_data.get_or_try_init(|| {
+            let datas = self.section_data.get_or_try_init_(|| {
                 let shdrs = self.ensure_shdrs()?;
                 let datas = (0..shdrs.len())
                     .map(|_| OnceCell::new())
@@ -433,7 +502,7 @@ where
             datas
                 .get(idx)
                 .ok_or_invalid_input(|| format!("ELF section index ({idx}) out of bounds"))?
-                .get_or_try_init(|| -> Result<Cow<'elf, [u8]>> {
+                .get_or_try_init_(|| -> Result<Cow<'elf, [u8]>> {
                     let data = self
                                 .backend
                                 .read_pod_slice::<u8>(shdr.offset(), shdr.size() as usize)
@@ -445,12 +514,12 @@ where
                         // data.
                         let (ch_type, ch_size) = if shdr.is_32bit() {
                             let chdr = data
-                                .read_pod_ref::<Elf32_Chdr>()
+                                .read_pod::<Elf32_Chdr>()
                                 .ok_or_invalid_data(|| "failed to read ELF compression header")?;
                             (chdr.ch_type, chdr.ch_size.into())
                         } else {
                             let chdr = data
-                                .read_pod_ref::<Elf64_Chdr>()
+                                .read_pod::<Elf64_Chdr>()
                                 .ok_or_invalid_data(|| "failed to read ELF compression header")?;
                             (chdr.ch_type, chdr.ch_size)
                         };
@@ -486,11 +555,11 @@ where
     fn read_first_shdr(&self, ehdr: &ElfN_Ehdr<'_>) -> Result<ElfN_Shdr<'elf>> {
         let shdr = if ehdr.is_32bit() {
             self.backend
-                .read_pod_obj::<Elf32_Shdr>(ehdr.shoff() as _)
+                .read_pod_obj::<Elf32_Shdr>(ehdr.shoff())
                 .map(ElfN_Shdr::B32)
         } else {
             self.backend
-                .read_pod_obj::<Elf64_Shdr>(ehdr.shoff() as _)
+                .read_pod_obj::<Elf64_Shdr>(ehdr.shoff())
                 .map(ElfN_Shdr::B64)
         }
         .context("failed to read ELF section header")?;
@@ -568,7 +637,7 @@ where
     }
 
     fn ensure_ehdr(&self) -> Result<&EhdrExt<'elf>> {
-        self.ehdr.get_or_try_init(|| self.parse_ehdr())
+        self.ehdr.get_or_try_init_(|| self.parse_ehdr())
     }
 
     fn parse_shdrs(&self) -> Result<ElfN_Shdrs<'elf>> {
@@ -590,7 +659,7 @@ where
     }
 
     fn ensure_shdrs(&self) -> Result<&ElfN_Shdrs<'elf>> {
-        self.shdrs.get_or_try_init(|| self.parse_shdrs())
+        self.shdrs.get_or_try_init_(|| self.parse_shdrs())
     }
 
     fn parse_phdrs(&self) -> Result<ElfN_Phdrs<'elf>> {
@@ -612,7 +681,7 @@ where
     }
 
     fn ensure_phdrs(&self) -> Result<&ElfN_Phdrs<'elf>> {
-        self.phdrs.get_or_try_init(|| self.parse_phdrs())
+        self.phdrs.get_or_try_init_(|| self.parse_phdrs())
     }
 
     fn shstrndx(&self, ehdr: &ElfN_Ehdr<'_>) -> Result<usize> {
@@ -642,7 +711,7 @@ where
     }
 
     fn ensure_shstrtab(&self) -> Result<&Cow<'elf, [u8]>> {
-        self.shstrtab.get_or_try_init(|| self.parse_shstrtab())
+        self.shstrtab.get_or_try_init_(|| self.parse_shstrtab())
     }
 
     /// Get the name of the section at a given index.
@@ -738,7 +807,7 @@ where
     }
 
     fn ensure_symtab_cache(&self) -> Result<&SymbolTableCache<'elf>> {
-        self.symtab.get_or_try_init(|| {
+        self.symtab.get_or_try_init_(|| {
             let syms = self.parse_syms(".symtab")?;
             let strtab = self.parse_strs(".strtab")?;
             let cache = SymbolTableCache::new(syms, strtab);
@@ -747,7 +816,7 @@ where
     }
 
     fn ensure_dynsym_cache(&self) -> Result<&SymbolTableCache<'elf>> {
-        self.dynsym.get_or_try_init(|| {
+        self.dynsym.get_or_try_init_(|| {
             // TODO: We really should check the `.dynamic` section for
             //       information on what symbol and string tables to
             //       use instead of hard coding names here.
@@ -824,6 +893,11 @@ impl Backend for Mmap {
     type ImplTy<'bcknd> = &'bcknd [u8];
 }
 
+impl Backend for StaticMem {
+    type ObjTy = StaticMem;
+    type ImplTy<'bcknd> = &'bcknd [u8];
+}
+
 impl Backend for File {
     type ObjTy = Box<File>;
     type ImplTy<'bcknd> = &'bcknd File;
@@ -859,7 +933,7 @@ impl<'elf> BackendImpl<'elf> for &'elf [u8] {
     {
         let value = self
             .get(offset as _..)
-            .unwrap()
+            .ok_or_invalid_data(|| "failed to read slice data: invalid offset")?
             .read_pod_slice_ref::<T>(count)
             .ok_or_invalid_data(|| "failed to read slice from mmap")?;
         Ok(Cow::Borrowed(value))
@@ -873,17 +947,8 @@ impl<'elf> BackendImpl<'elf> for &File {
     {
         let mut slf = *self;
         let _pos = slf.seek(SeekFrom::Start(offset))?;
-
-        let mut value = MaybeUninit::<T>::zeroed();
-        // SAFETY: `value` is a buffer of `size_of::<T>` bytes.
-        // TODO: Are we UB here because we create a mut reference?
-        let slice = unsafe {
-            slice::from_raw_parts_mut(value.as_mut_ptr().cast::<u8>(), mem::size_of::<T>())
-        };
-        let () = slf.read_exact(slice)?;
-        // SAFETY: `T` is a `Pod` and hence valid for any bit pattern,
-        //         including all zeroes.
-        Ok(Cow::Owned(unsafe { value.assume_init() }))
+        let obj = read_pod::<T, _>(&mut slf).map(Cow::Owned)?;
+        Ok(obj)
     }
 
     fn read_pod_slice<T>(&self, offset: u64, count: usize) -> Result<Cow<'elf, [T]>, Error>
@@ -892,17 +957,8 @@ impl<'elf> BackendImpl<'elf> for &File {
     {
         let mut slf = *self;
         let _pos = slf.seek(SeekFrom::Start(offset))?;
-        let mut vec = Vec::<T>::new();
-        // SAFETY: `T` is a `Pod` and hence valid for any bit pattern,
-        //         including all zeroes.
-        let () = vec.resize(count, unsafe { mem::zeroed() });
-
-        // TODO: Are we UB here because we create a mut reference?
-        let slice = unsafe {
-            slice::from_raw_parts_mut(vec.as_mut_ptr().cast::<u8>(), count * mem::size_of::<T>())
-        };
-        let () = slf.read_exact(slice)?;
-        Ok(Cow::Owned(vec))
+        let vec = read_pod_vec::<T, _>(&mut slf, count).map(Cow::Owned)?;
+        Ok(vec)
     }
 }
 
@@ -981,7 +1037,7 @@ impl ElfParser<Mmap> {
         let data = unsafe { mem::transmute::<&[u8], &'static [u8]>(mmap.deref()) };
         let _backend = mmap;
 
-        let parser = ElfParser {
+        let parser = Self {
             cache: Cache::new(data),
             module,
             _backend,
@@ -1006,6 +1062,18 @@ impl ElfParser<Mmap> {
     }
 }
 
+impl ElfParser<StaticMem> {
+    /// Create an `ElfParser` from a region of static memory.
+    pub(crate) fn from_mem(mem: StaticMem) -> Self {
+        Self {
+            cache: Cache::new(mem),
+            // TODO: Should provide the module.
+            module: None,
+            _backend: mem,
+        }
+    }
+}
+
 impl<B> ElfParser<B>
 where
     B: Backend,
@@ -1013,7 +1081,9 @@ where
     /// Retrieve the data corresponding to the ELF section at index
     /// `idx`, optionally decompressing it if it is compressed.
     pub(crate) fn section_data(&self, idx: usize) -> Result<&[u8]> {
-        self.cache.section_data(idx)
+        self.cache
+            .section_data(idx)
+            .with_context(|| format!("failed to read ELF section with index {idx}"))
     }
 
     /// Find the section of a given name.
@@ -1024,53 +1094,6 @@ where
         Ok(index)
     }
 
-    pub(crate) fn find_sym(
-        &self,
-        addr: Addr,
-        opts: &FindSymOpts,
-    ) -> Result<Result<ResolvedSym<'_>, Reason>> {
-        // ELF doesn't carry any source code or inlining information.
-        let _opts = opts;
-
-        let symtab_cache = self.cache.ensure_symtab_cache()?;
-        let symtab_by_addr_idx = symtab_cache.ensure_by_addr_idx();
-
-        if let Some(sym) = find_sym(
-            &symtab_cache.syms,
-            self.module.as_deref(),
-            symtab_by_addr_idx,
-            &symtab_cache.strs,
-            addr,
-            SymType::Undefined,
-        )? {
-            return Ok(Ok(sym))
-        }
-
-        let dynsym_cache = self.cache.ensure_dynsym_cache()?;
-        let dynsym_by_addr_idx = dynsym_cache.ensure_by_addr_idx();
-        if let Some(sym) = find_sym(
-            &dynsym_cache.syms,
-            self.module.as_deref(),
-            dynsym_by_addr_idx,
-            &dynsym_cache.strs,
-            addr,
-            SymType::Undefined,
-        )? {
-            return Ok(Ok(sym))
-        }
-
-        // At this point we haven't found a symbol for the given
-        // address. The emptiness of `dynsym` has no bearing on the
-        // reason we report -- for all intents and purposes it is either
-        // required or not at all necessary.
-        let reason = if symtab_cache.syms.is_empty() {
-            Reason::MissingSyms
-        } else {
-            Reason::UnknownAddr
-        };
-        Ok(Err(reason))
-    }
-
     fn find_addr_impl<'slf>(
         &'slf self,
         name: &str,
@@ -1097,7 +1120,7 @@ where
                     if sym.st_shndx != SHN_UNDEF {
                         found.push(SymInfo {
                             name: Cow::Borrowed(name_visit.name(strs)?),
-                            addr: sym.st_value as Addr,
+                            addr: sym.st_value,
                             size: Some(sym.st_size as usize),
                             // SANITY: We filter out all unsupported symbol
                             //         types, so this conversion should always
@@ -1109,6 +1132,7 @@ where
                                 .transpose()?
                                 .flatten(),
                             module: self.module.as_deref().map(Cow::Borrowed),
+                            _non_exhaustive: (),
                         });
                     }
                 }
@@ -1160,7 +1184,7 @@ where
             if sym.matches(opts.sym_type) && sym.st_shndx != SHN_UNDEF {
                 let sym_info = SymInfo {
                     name: Cow::Borrowed(name.name(strs)?),
-                    addr: sym.st_value as Addr,
+                    addr: sym.st_value,
                     size: Some(sym.st_size as usize),
                     // SANITY: We filter out all unsupported symbol
                     //         types, so this conversion should always
@@ -1172,6 +1196,7 @@ where
                         .transpose()?
                         .flatten(),
                     module: None,
+                    _non_exhaustive: (),
                 };
                 if let ControlFlow::Break(()) = f(&sym_info) {
                     return Ok(())
@@ -1203,13 +1228,14 @@ where
     /// Find the file offset of the symbol at address `addr`.
     // If possible, use the constant-time [`file_offset`] function
     // instead.
-    pub(crate) fn find_file_offset(&self, addr: Addr) -> Result<Option<u64>> {
+    pub(crate) fn find_file_offset(&self, addr: Addr, size: usize) -> Result<Option<u64>> {
         let phdrs = self.program_headers()?;
         let offset = phdrs.iter(0).find_map(|phdr| {
             let phdr = phdr.to_64bit();
 
             if phdr.p_type == PT_LOAD {
-                if (phdr.p_vaddr..phdr.p_vaddr + phdr.p_memsz).contains(&addr) {
+                let limit = phdr.p_vaddr + phdr.p_filesz;
+                if (phdr.p_vaddr..limit).contains(&addr) || (addr == limit && size == 0) {
                     return Some(addr - phdr.p_vaddr + phdr.p_offset)
                 }
             }
@@ -1243,7 +1269,7 @@ where
             let phdr = phdr.to_64bit();
             if phdr.p_type == PT_LOAD {
                 if (phdr.p_offset..phdr.p_offset + phdr.p_filesz).contains(&offset) {
-                    return Some((offset - phdr.p_offset + phdr.p_vaddr) as Addr)
+                    return Some(offset - phdr.p_offset + phdr.p_vaddr)
                 }
             }
             None
@@ -1284,11 +1310,7 @@ where
 
         let idx = by_addr_idx.get(idx).unwrap();
         let sym_name = self.get_symbol_name(*idx).unwrap();
-        (
-            sym_name,
-            addr as Addr,
-            usize::try_from(size).unwrap_or(usize::MAX),
-        )
+        (sym_name, addr, usize::try_from(size).unwrap_or(usize::MAX))
     }
 
     pub(crate) fn cache(&self) -> Result<()> {
@@ -1306,7 +1328,61 @@ where
     }
 }
 
-impl Debug for ElfParser {
+// TODO: Ideally we wouldn't have to implement this trait for this type,
+//       but right now we need it because `ElfResolver` is not generic
+//       over the backend in use.
+impl<B> Symbolize for ElfParser<B>
+where
+    B: Backend,
+{
+    fn find_sym(&self, addr: Addr, opts: &FindSymOpts) -> Result<Result<ResolvedSym<'_>, Reason>> {
+        // ELF doesn't carry any source code or inlining information.
+        let _opts = opts;
+
+        let symtab_cache = self.cache.ensure_symtab_cache()?;
+        let symtab_by_addr_idx = symtab_cache.ensure_by_addr_idx();
+
+        if let Some(sym) = find_sym(
+            &symtab_cache.syms,
+            self.module.as_deref(),
+            symtab_by_addr_idx,
+            &symtab_cache.strs,
+            addr,
+            SymType::Undefined,
+        )? {
+            return Ok(Ok(sym))
+        }
+
+        let dynsym_cache = self.cache.ensure_dynsym_cache()?;
+        let dynsym_by_addr_idx = dynsym_cache.ensure_by_addr_idx();
+        if let Some(sym) = find_sym(
+            &dynsym_cache.syms,
+            self.module.as_deref(),
+            dynsym_by_addr_idx,
+            &dynsym_cache.strs,
+            addr,
+            SymType::Undefined,
+        )? {
+            return Ok(Ok(sym))
+        }
+
+        // At this point we haven't found a symbol for the given
+        // address. The emptiness of `dynsym` has no bearing on the
+        // reason we report -- for all intents and purposes it is either
+        // required or not at all necessary.
+        let reason = if symtab_cache.syms.is_empty() {
+            Reason::MissingSyms
+        } else {
+            Reason::UnknownAddr
+        };
+        Ok(Err(reason))
+    }
+}
+
+impl<B> Debug for ElfParser<B>
+where
+    B: Backend,
+{
     fn fmt(&self, f: &mut Formatter<'_>) -> FmtResult {
         let module = self
             .module
@@ -1323,7 +1399,6 @@ mod tests {
 
     use super::super::types::SHN_LORESERVE;
 
-    use std::env;
     use std::env::current_exe;
     #[cfg(feature = "nightly")]
     use std::hint::black_box;
@@ -1332,9 +1407,12 @@ mod tests {
     use std::path::Path;
     use std::slice;
 
+    use miniz_oxide::deflate::compress_to_vec_zlib;
+
     use tempfile::NamedTempFile;
 
     use test_log::test;
+    use test_tag::tag;
 
     #[cfg(feature = "nightly")]
     use test::Bencher;
@@ -1374,11 +1452,36 @@ mod tests {
         assert_ne!(format!("{parser:?}"), "");
     }
 
+    /// Check that we can read Pod style objects from a reader without
+    /// triggering undefined behavior.
+    #[tag(miri)]
+    #[test]
+    fn pod_obj_reading() {
+        #[repr(C)]
+        #[derive(Clone, Debug)]
+        struct Foo {
+            x: u32,
+            y: u8,
+            z: u64,
+        }
+
+        unsafe impl Pod for Foo {}
+
+        let buf = [42u8; mem::size_of::<Foo>()];
+        let foo = read_pod::<Foo, _>(&mut buf.as_slice()).unwrap();
+        assert_eq!(foo.y, 42);
+
+        let buf = [42u8; 2 * mem::size_of::<Foo>()];
+        let vec = read_pod_vec::<Foo, _>(&mut buf.as_slice(), 2).unwrap();
+        assert_eq!(vec[0].y, 42);
+        assert_eq!(vec[1].y, 42);
+    }
+
     /// Check that our `ElfParser` can handle more than 0xff00 section
     /// headers and more than 0xffff program headers properly.
     #[test]
     fn excessive_section_and_program_headers() {
-        const SHNUM: u16 = (SHN_LORESERVE + 0x42) as _;
+        const SHNUM: u16 = SHN_LORESERVE + 0x42;
         const PHNUM: u32 = 0xffff + 0x43;
 
         #[repr(C)]
@@ -1461,7 +1564,7 @@ mod tests {
     /// 0xff00.
     #[test]
     fn large_e_shstrndx() {
-        const SHSTRNDX: u16 = (SHN_LORESERVE + 0x42) as _;
+        const SHSTRNDX: u16 = SHN_LORESERVE + 0x42;
 
         #[repr(C)]
         struct Elf {
@@ -1585,8 +1688,10 @@ mod tests {
             let parser = ElfParser::open(path).unwrap();
             let () = parser
                 .for_each(&opts, &mut |sym| {
-                    let file_offset = parser.find_file_offset(sym.addr).unwrap();
-                    assert_eq!(file_offset, sym.file_offset);
+                    let file_offset = parser
+                        .find_file_offset(sym.addr, sym.size.unwrap())
+                        .unwrap();
+                    assert_eq!(sym.file_offset, file_offset, "{sym:#x?}");
                     ControlFlow::Continue(())
                 })
                 .unwrap();
@@ -1949,6 +2054,110 @@ mod tests {
         assert!(symtab.is_empty());
     }
 
+    /// Check that we can decompress ELF section data with an unaligned
+    /// compression header.
+    #[test]
+    fn unaligned_compression_header_reading() {
+        let ehdr = Elf64_Ehdr {
+            e_ident: [127, 69, 76, 70, 2, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0],
+            e_type: 3,
+            e_machine: 62,
+            e_version: 1,
+            e_entry: 0,
+            e_phoff: 0,
+            e_shoff: 0,
+            e_flags: 0,
+            e_ehsize: 0,
+            e_phentsize: 0,
+            e_phnum: 0,
+            e_shentsize: 0,
+            e_shnum: 2,
+            e_shstrndx: 1,
+        };
+        let ehdr = EhdrExt {
+            ehdr: ElfN_Ehdr::B64(Cow::Borrowed(&ehdr)),
+            shnum: 2,
+            phnum: 0,
+        };
+
+        let data = [];
+        let zlib_hdr = compress_to_vec_zlib(&data, 0);
+        let chdr = Elf64_Chdr {
+            ch_type: ELFCOMPRESS_ZLIB,
+            ch_reserved: 0,
+            ch_size: 0,
+            ch_addralign: 1,
+        };
+        let shdrs = [
+            Elf64_Shdr {
+                sh_name: 0,
+                sh_type: 0,
+                sh_flags: 0,
+                sh_addr: 0,
+                sh_offset: 0,
+                sh_size: 0,
+                sh_link: 0,
+                sh_info: 0,
+                sh_addralign: 0,
+                sh_entsize: 0,
+            },
+            Elf64_Shdr {
+                sh_name: 0,
+                sh_type: 0,
+                sh_flags: SHF_COMPRESSED,
+                sh_addr: 0,
+                sh_offset: 0,
+                sh_size: (mem::size_of_val(&chdr) + zlib_hdr.len()) as _,
+                sh_link: 0,
+                sh_info: 0,
+                sh_addralign: 0,
+                sh_entsize: 0,
+            },
+        ];
+
+        let mut aligned_data = [0u64; 128];
+
+        // Write a compression header at an unaligned address, followed
+        // by the zlib data.
+        let () = unsafe {
+            aligned_data
+                .as_mut_ptr()
+                .cast::<u8>()
+                .add(1)
+                .cast::<Elf64_Chdr>()
+                .write_unaligned(chdr)
+        };
+        let () = unsafe {
+            aligned_data
+                .as_mut_ptr()
+                .cast::<u8>()
+                .add(1)
+                .add(mem::size_of::<Elf64_Chdr>())
+                .copy_from(zlib_hdr.as_ptr(), zlib_hdr.len())
+        };
+        let aligned_data = unsafe {
+            slice::from_raw_parts(
+                aligned_data.as_ptr().cast::<u8>(),
+                aligned_data.len() * size_of_val(&aligned_data[0]),
+            )
+        };
+        let unaligned_data = &aligned_data[1..];
+
+        let cache = Cache {
+            backend: unaligned_data,
+            ehdr: OnceCell::from(ehdr),
+            shdrs: OnceCell::from(ElfN_Shdrs::B64(Cow::Borrowed(shdrs.as_slice()))),
+            shstrtab: OnceCell::from(Cow::Borrowed(b".debug_info\x00".as_slice())),
+            phdrs: OnceCell::new(),
+            symtab: OnceCell::new(),
+            dynsym: OnceCell::new(),
+            section_data: OnceCell::from(vec![OnceCell::new(), OnceCell::new()].into_boxed_slice()),
+        };
+
+        let new_data = cache.section_data(1).unwrap();
+        assert_eq!(new_data, data);
+    }
+
     /// Benchmark creation of our "str2symtab" table.
     ///
     /// Creating this table exercises a lot of the parser code paths and
