@@ -1,4 +1,5 @@
 use std::borrow::Cow;
+use std::cell::OnceCell;
 use std::fmt::Debug;
 use std::fmt::Formatter;
 use std::fmt::Result as FmtResult;
@@ -16,7 +17,6 @@ use crate::inspect::FindAddrOpts;
 use crate::inspect::ForEachFn;
 use crate::inspect::Inspect;
 use crate::inspect::SymInfo;
-use crate::once::OnceCell;
 use crate::symbolize::FindSymOpts;
 use crate::symbolize::Reason;
 use crate::symbolize::ResolvedSym;
@@ -162,6 +162,7 @@ impl<'kfunc> TryFrom<&'kfunc Kfunc> for SymInfo<'kfunc> {
             sym_type: SymType::Function,
             file_offset: None,
             module: None,
+            _non_exhaustive: (),
         };
         Ok(sym)
     }
@@ -222,7 +223,7 @@ impl KsymResolver {
             }
         }
 
-        let () = syms.sort_by_key(|a| a.addr());
+        let () = syms.sort_by_key(Ksym::addr);
 
         let slf = Self {
             syms: syms.into_boxed_slice(),
@@ -402,7 +403,7 @@ ffffffffc212d000 t ftrace_trampoline    [__builtin__ftrace]
 
             let prog = resolver.syms[1].as_bpf_prog().unwrap();
             assert_eq!(prog.addr(), 0xffffffffc003e9c8);
-            assert_eq!(prog.name(), "kprobe__cap_capable");
+            assert_eq!(prog.name(), "bpf_prog_30304e82b4033ea3_kprobe__cap_capable");
             assert_eq!(
                 prog.tag(),
                 BpfTag::from([0x30, 0x30, 0x4e, 0x82, 0xb4, 0x03, 0x3e, 0xa3])
