@@ -1,4 +1,5 @@
 use std::borrow::Cow;
+use std::cell::OnceCell;
 use std::cell::RefCell;
 use std::collections::HashMap;
 use std::ffi::OsStr;
@@ -20,6 +21,7 @@ use crate::breakpad::BreakpadResolver;
 use crate::elf::ElfParser;
 use crate::elf::ElfResolver;
 use crate::elf::ElfResolverData;
+use crate::elf::StaticMem;
 #[cfg(feature = "dwarf")]
 use crate::elf::DEFAULT_DEBUG_DIRS;
 use crate::file_cache::FileCache;
@@ -38,8 +40,9 @@ use crate::mmap::Mmap;
 use crate::normalize;
 use crate::normalize::normalize_sorted_user_addrs_with_entries;
 use crate::normalize::Handler as _;
+#[cfg(feature = "apk")]
+use crate::pathlike::PathLike;
 use crate::perf_map::PerfMap;
-use crate::symbolize::InlinedFn;
 use crate::symbolize::Resolve;
 use crate::symbolize::TranslateFileOffset;
 use crate::util;
@@ -48,6 +51,9 @@ use crate::util::uname_release;
 use crate::util::Dbg;
 #[cfg(feature = "tracing")]
 use crate::util::Hexify;
+use crate::util::OnceCellExt as _;
+use crate::vdso::create_vdso_parser;
+use crate::vdso::VDSO_MAPS_COMPONENT;
 #[cfg(feature = "apk")]
 use crate::zip;
 use crate::Addr;
@@ -118,7 +124,7 @@ impl Debug for DebugMapsEntry<'_> {
 
 /// Demangle a symbol name using the demangling scheme for the given language.
 #[cfg(feature = "demangle")]
-fn maybe_demangle(name: Cow<'_, str>, language: SrcLang) -> Cow<'_, str> {
+fn maybe_demangle_impl(name: Cow<'_, str>, language: SrcLang) -> Cow<'_, str> {
     match language {
         SrcLang::Rust => rustc_demangle::try_demangle(name.as_ref())
             .ok()
@@ -140,11 +146,72 @@ fn maybe_demangle(name: Cow<'_, str>, language: SrcLang) -> Cow<'_, str> {
 }
 
 #[cfg(not(feature = "demangle"))]
-fn maybe_demangle(name: Cow<'_, str>, _language: SrcLang) -> Cow<'_, str> {
+fn maybe_demangle_impl(name: Cow<'_, str>, _language: SrcLang) -> Cow<'_, str> {
     // Demangling is disabled.
     name
 }
 
+/// Demangle the provided symbol if asked for and possible.
+fn maybe_demangle(symbol: Cow<'_, str>, language: SrcLang, demangle: bool) -> Cow<'_, str> {
+    if demangle {
+        maybe_demangle_impl(symbol, language)
+    } else {
+        symbol
+    }
+}
+
+
+/// Symbolize an address using the provided [`Resolver`].
+pub(crate) fn symbolize_with_resolver<'slf>(
+    addr: Addr,
+    resolver: &Resolver<'_, 'slf>,
+    find_sym_opts: &FindSymOpts,
+    demangle: bool,
+) -> Result<Symbolized<'slf>> {
+    /// Convert a `ResolvedSym` into a `Sym`, potentially performing
+    /// symbol name demangling in the process.
+    fn convert_sym<'sym>(addr: Addr, sym: ResolvedSym<'sym>, demangle: bool) -> Sym<'sym> {
+        let ResolvedSym {
+            name,
+            module,
+            addr: sym_addr,
+            size,
+            lang,
+            code_info,
+            mut inlined,
+        } = sym;
+
+        let () = inlined.iter_mut().for_each(|inlined_fn| {
+            let name = take(&mut inlined_fn.name);
+            inlined_fn.name = maybe_demangle(name, lang, demangle);
+        });
+
+        let sym = Sym {
+            name: maybe_demangle(Cow::Borrowed(name), lang, demangle),
+            module: module.map(Cow::Borrowed),
+            addr: sym_addr,
+            offset: (addr - sym_addr) as usize,
+            size,
+            code_info,
+            inlined,
+            _non_exhaustive: (),
+        };
+        sym
+    }
+
+    let sym = match resolver {
+        Resolver::Uncached(resolver) => match resolver.find_sym(addr, find_sym_opts)? {
+            Ok(sym) => convert_sym(addr, sym, demangle).into_owned(),
+            Err(reason) => return Ok(Symbolized::Unknown(reason)),
+        },
+        Resolver::Cached(resolver) => match resolver.find_sym(addr, find_sym_opts)? {
+            Ok(sym) => convert_sym(addr, sym, demangle),
+            Err(reason) => return Ok(Symbolized::Unknown(reason)),
+        },
+    };
+
+    Ok(Symbolized::Sym(sym))
+}
 
 /// Information about a member inside an APK.
 ///
@@ -196,12 +263,13 @@ impl<F> ProcessDispatch for F where F: Fn(ProcessMemberInfo<'_>) -> Result<Optio
 
 #[cfg(feature = "apk")]
 fn default_apk_dispatcher(
+    apk_path: &dyn PathLike,
     info: ApkMemberInfo<'_>,
     debug_dirs: Option<&[PathBuf]>,
 ) -> Result<Box<dyn Resolve>> {
     // Create an Android-style binary-in-APK path for
     // reporting purposes.
-    let apk_elf_path = create_apk_elf_path(info.apk_path, info.member_path);
+    let apk_elf_path = create_apk_elf_path(apk_path.represented_path(), info.member_path);
     let parser = Rc::new(ElfParser::from_mmap(
         info.member_mmap,
         Some(apk_elf_path.into_os_string()),
@@ -268,6 +336,11 @@ impl Builder {
 
     /// Enable/disable source code location information (line numbers,
     /// file names etc.).
+    ///
+    /// This option only has an effect if `debug_syms` of the particular
+    /// symbol source is set to `true`. Furthermore, it is a necessary
+    /// prerequisite for retrieving inlined function information (see
+    /// [`Self::enable_inlined_fns`]).
     pub fn enable_code_info(mut self, enable: bool) -> Self {
         self.code_info = enable;
         self
@@ -384,6 +457,7 @@ impl Builder {
             perf_map_cache: FileCache::builder().enable_auto_reload(auto_reload).build(),
             process_vma_cache: RefCell::new(HashMap::new()),
             process_cache: InsertMap::new(),
+            vdso_parser: OnceCell::new(),
             find_sym_opts,
             demangle,
             #[cfg(feature = "dwarf")]
@@ -429,6 +503,9 @@ struct SymbolizeHandler<'sym> {
     /// Whether to work with `/proc/<pid>/map_files/` entries or with
     /// symbolic paths mentioned in `/proc/<pid>/maps` instead.
     map_files: bool,
+    /// Whether or not to symbolize addresses in a vDSO (virtual dynamic
+    /// shared object).
+    vdso: bool,
     /// Symbols representing the symbolized addresses.
     all_symbols: Vec<Symbolized<'sym>>,
 }
@@ -436,16 +513,16 @@ struct SymbolizeHandler<'sym> {
 impl SymbolizeHandler<'_> {
     #[cfg(feature = "apk")]
     fn handle_apk_addr(&mut self, addr: Addr, file_off: u64, entry_path: &EntryPath) -> Result<()> {
-        let apk_path = if self.map_files {
-            &entry_path.maps_file
+        let result = if self.map_files {
+            self.symbolizer
+                .apk_resolver(entry_path, file_off, self.debug_syms)?
         } else {
-            &entry_path.symbolic_path
+            let path = &entry_path.symbolic_path;
+            self.symbolizer
+                .apk_resolver(path, file_off, self.debug_syms)?
         };
 
-        match self
-            .symbolizer
-            .apk_resolver(apk_path, file_off, self.debug_syms)?
-        {
+        match result {
             Some((elf_resolver, elf_addr)) => {
                 let symbol = self.symbolizer.symbolize_with_resolver(
                     elf_addr,
@@ -465,7 +542,7 @@ impl SymbolizeHandler<'_> {
                 self.symbolizer.maybe_debug_dirs(self.debug_syms),
             )
         } else {
-            let path = entry_path.symbolic_path.as_path();
+            let path = &entry_path.symbolic_path;
             self.symbolizer
                 .elf_cache
                 .elf_resolver(path, self.symbolizer.maybe_debug_dirs(self.debug_syms))
@@ -485,7 +562,7 @@ impl SymbolizeHandler<'_> {
     }
 
     fn handle_perf_map_addr(&mut self, addr: Addr) -> Result<()> {
-        if let Some(perf_map) = self.symbolizer.perf_map(self.pid)? {
+        if let Some(perf_map) = self.symbolizer.perf_map_resolver(self.pid)? {
             let symbolized = self
                 .symbolizer
                 .symbolize_with_resolver(addr, &Resolver::Cached(perf_map))?;
@@ -495,6 +572,25 @@ impl SymbolizeHandler<'_> {
         }
         Ok(())
     }
+
+    fn handle_vdso_addr(
+        &mut self,
+        addr: Addr,
+        file_off: u64,
+        vdso_range: &Range<Addr>,
+    ) -> Result<()> {
+        let parser = self.symbolizer.vdso_parser(self.pid, vdso_range)?;
+        match parser.file_offset_to_virt_offset(file_off)? {
+            Some(addr) => {
+                let symbol = self
+                    .symbolizer
+                    .symbolize_with_resolver(addr, &Resolver::Cached(parser))?;
+                let () = self.all_symbols.push(symbol);
+            }
+            None => self.handle_unknown_addr(addr, Reason::InvalidFileOffset),
+        }
+        Ok(())
+    }
 }
 
 impl normalize::Handler<Reason> for SymbolizeHandler<'_> {
@@ -505,12 +601,13 @@ impl normalize::Handler<Reason> for SymbolizeHandler<'_> {
 
     #[cfg_attr(feature = "tracing", crate::log::instrument(skip_all, fields(addr = format_args!("{addr:#x}"), entry = ?DebugMapsEntry(entry))))]
     fn handle_entry_addr(&mut self, addr: Addr, entry: &MapsEntry) -> Result<()> {
+        let file_off = addr - entry.range.start + entry.offset;
+
         if let Some(path_name) = &entry.path_name {
             if let Some(resolver) = self
                 .symbolizer
                 .process_dispatch_resolver(entry.range.clone(), path_name)?
             {
-                let file_off = addr - entry.range.start + entry.offset;
                 let () = match resolver.file_offset_to_virt_offset(file_off)? {
                     Some(addr) => {
                         let symbol = self.symbolizer.symbolize_with_resolver(
@@ -531,7 +628,6 @@ impl normalize::Handler<Reason> for SymbolizeHandler<'_> {
 
         match &entry.path_name {
             Some(PathName::Path(entry_path)) => {
-                let file_off = addr - entry.range.start + entry.offset;
                 let ext = entry_path
                     .symbolic_path
                     .extension()
@@ -542,8 +638,15 @@ impl normalize::Handler<Reason> for SymbolizeHandler<'_> {
                     _ => self.handle_elf_addr(addr, file_off, entry_path),
                 }
             }
-            Some(PathName::Component(..)) => {
-                let () = self.handle_unknown_addr(addr, Reason::Unsupported);
+            Some(PathName::Component(component)) => {
+                match component.as_str() {
+                    component if self.vdso && component == VDSO_MAPS_COMPONENT => {
+                        let () = self.handle_vdso_addr(addr, file_off, &entry.range)?;
+                    }
+                    _ => {
+                        let () = self.handle_unknown_addr(addr, Reason::Unsupported);
+                    }
+                }
                 Ok(())
             }
             // If there is no path associated with this entry, we don't
@@ -574,9 +677,9 @@ impl normalize::Handler<Reason> for SymbolizeHandler<'_> {
 ///
 /// Objects of this type are at the core of our logic determining whether to
 /// heap allocate certain data such as paths or symbol names or whether to just
-/// hand out references to mmap'ed data.
+/// hand out references to mmap'ed (or potentially static) data.
 #[derive(Debug)]
-enum Resolver<'tmp, 'slf> {
+pub(crate) enum Resolver<'tmp, 'slf> {
     Uncached(&'tmp (dyn Symbolize + 'tmp)),
     Cached(&'slf dyn Symbolize),
 }
@@ -622,6 +725,8 @@ pub struct Symbolizer {
     /// data by the user.
     process_vma_cache: RefCell<HashMap<Pid, Box<[maps::MapsEntry]>>>,
     process_cache: InsertMap<PathName, Option<Box<dyn Resolve>>>,
+    /// The ELF parser used for the system-wide vDSO.
+    vdso_parser: OnceCell<Box<ElfParser<StaticMem>>>,
     find_sym_opts: FindSymOpts,
     demangle: bool,
     #[cfg(feature = "dwarf")]
@@ -643,98 +748,34 @@ impl Symbolizer {
         Builder::default()
     }
 
-    /// Demangle the provided symbol if asked for and possible.
-    fn maybe_demangle<'sym>(&self, symbol: Cow<'sym, str>, language: SrcLang) -> Cow<'sym, str> {
-        if self.demangle {
-            maybe_demangle(symbol, language)
-        } else {
-            symbol
-        }
+    /// Register an [`ElfResolver`] to use for subsequent symbolization
+    /// requests.
+    ///
+    /// Register an existing externally managed [`ElfResolver`] object
+    /// to use in subsequent symbolization requests involving `path`.
+    /// Doing so allows for reuse of already parsed ELF data.
+    ///
+    /// This method will fail if a cached [`ElfResolver`] is already
+    /// present for the given path.
+    pub fn register_elf_resolver(
+        &mut self,
+        path: &Path,
+        elf_resolver: Rc<ElfResolver>,
+    ) -> Result<()> {
+        self.elf_cache.register(path, elf_resolver)
     }
 
-    /// Symbolize an address using the provided [`SymResolver`].
+    /// Symbolize an address using the provided [`Resolver`].
     #[cfg_attr(feature = "tracing", crate::log::instrument(skip_all, fields(addr = format_args!("{addr:#x}"), resolver = ?resolver.inner())))]
     fn symbolize_with_resolver<'slf>(
         &'slf self,
         addr: Addr,
         resolver: &Resolver<'_, 'slf>,
     ) -> Result<Symbolized<'slf>> {
-        let (sym_name, sym_module, sym_addr, sym_size, code_info, inlined) = match resolver {
-            Resolver::Uncached(resolver) => match resolver.find_sym(addr, &self.find_sym_opts)? {
-                Ok(sym) => {
-                    let ResolvedSym {
-                        name,
-                        module,
-                        addr,
-                        size,
-                        lang,
-                        code_info,
-                        inlined,
-                    } = sym;
-
-                    let name =
-                        Cow::Owned(self.maybe_demangle(Cow::Borrowed(name), lang).into_owned());
-                    let module = module.map(|module| Cow::Owned(module.to_os_string()));
-                    let code_info = code_info.map(|info| info.to_owned());
-                    let inlined = Vec::from(inlined)
-                        .into_iter()
-                        .map(|inlined_fn| {
-                            let InlinedFn {
-                                name,
-                                code_info,
-                                _non_exhaustive: (),
-                            } = inlined_fn;
-                            InlinedFn {
-                                name: Cow::Owned(self.maybe_demangle(name, lang).into_owned()),
-                                code_info: code_info.map(|info| info.to_owned()),
-                                _non_exhaustive: (),
-                            }
-                        })
-                        .collect::<Vec<_>>()
-                        .into_boxed_slice();
-
-                    (name, module, addr, size, code_info, inlined)
-                }
-                Err(reason) => return Ok(Symbolized::Unknown(reason)),
-            },
-            Resolver::Cached(resolver) => match resolver.find_sym(addr, &self.find_sym_opts)? {
-                Ok(sym) => {
-                    let ResolvedSym {
-                        name,
-                        module,
-                        addr,
-                        size,
-                        lang,
-                        code_info,
-                        mut inlined,
-                    } = sym;
-
-                    let name = self.maybe_demangle(Cow::Borrowed(name), lang);
-                    let module = module.map(Cow::Borrowed);
-                    let () = inlined.iter_mut().for_each(|inlined_fn| {
-                        let name = take(&mut inlined_fn.name);
-                        inlined_fn.name = self.maybe_demangle(name, lang);
-                    });
-                    (name, module, addr, size, code_info, inlined)
-                }
-                Err(reason) => return Ok(Symbolized::Unknown(reason)),
-            },
-        };
-
-        let sym = Sym {
-            name: sym_name,
-            module: sym_module,
-            addr: sym_addr,
-            offset: (addr - sym_addr) as usize,
-            size: sym_size,
-            code_info,
-            inlined,
-            _non_exhaustive: (),
-        };
-        Ok(Symbolized::Sym(sym))
+        symbolize_with_resolver(addr, resolver, &self.find_sym_opts, self.demangle)
     }
 
-    /// Symbolize a list of addresses using the provided [`SymResolver`].
+    /// Symbolize a list of addresses using the provided [`Resolver`].
     fn symbolize_addrs<'slf>(
         &'slf self,
         addrs: &[Addr],
@@ -755,7 +796,7 @@ impl Symbolizer {
     #[cfg(feature = "gsym")]
     fn gsym_resolver<'slf>(&'slf self, path: &Path) -> Result<&'slf GsymResolver<'static>> {
         let (file, cell) = self.gsym_cache.entry(path)?;
-        let resolver = cell.get_or_try_init(|| self.create_gsym_resolver(path, file))?;
+        let resolver = cell.get_or_try_init_(|| self.create_gsym_resolver(path, file))?;
         Ok(resolver)
     }
 
@@ -763,15 +804,17 @@ impl Symbolizer {
     fn create_apk_resolver<'slf>(
         &'slf self,
         apk: &zip::Archive,
-        apk_path: &Path,
+        apk_path: &dyn PathLike,
         file_off: u64,
         debug_dirs: Option<&[PathBuf]>,
         resolver_map: &'slf InsertMap<Range<u64>, Box<dyn Resolve>>,
     ) -> Result<Option<(&'slf dyn Resolve, Addr)>> {
+        let actual_path = apk_path.actual_path();
         // Find the APK entry covering the calculated file offset.
         for apk_entry in apk.entries() {
-            let apk_entry = apk_entry
-                .with_context(|| format!("failed to iterate `{}` members", apk_path.display()))?;
+            let apk_entry = apk_entry.with_context(|| {
+                format!("failed to iterate `{}` members", actual_path.display())
+            })?;
             let bounds = apk_entry.data_offset..apk_entry.data_offset + apk_entry.data.len() as u64;
 
             if bounds.contains(&file_off) {
@@ -782,11 +825,11 @@ impl Symbolizer {
                         .ok_or_invalid_input(|| {
                             format!(
                                 "invalid APK entry data bounds ({bounds:?}) in {}",
-                                apk_path.display()
+                                actual_path.display()
                             )
                         })?;
                     let info = ApkMemberInfo {
-                        apk_path,
+                        apk_path: actual_path,
                         member_path: apk_entry.path,
                         member_mmap: mmap,
                         _non_exhaustive: (),
@@ -796,10 +839,10 @@ impl Symbolizer {
                         if let Some(resolver) = (apk_dispatch)(info.clone())? {
                             resolver
                         } else {
-                            default_apk_dispatcher(info, debug_dirs)?
+                            default_apk_dispatcher(apk_path, info, debug_dirs)?
                         }
                     } else {
-                        default_apk_dispatcher(info, debug_dirs)?
+                        default_apk_dispatcher(apk_path, info, debug_dirs)?
                     };
 
                     Ok(resolver)
@@ -816,22 +859,21 @@ impl Symbolizer {
         Ok(None)
     }
 
-    // TODO: Should likely require a `PathLike` as input, similar to
-    //       `FileCache<ElfResolverData>::elf_resolver`.
     #[cfg(feature = "apk")]
     fn apk_resolver<'slf>(
         &'slf self,
-        path: &Path,
+        path: &dyn PathLike,
         file_off: u64,
         debug_syms: bool,
     ) -> Result<Option<(&'slf dyn Resolve, Addr)>> {
-        let (file, cell) = self.apk_cache.entry(path)?;
-        let (apk, resolvers) = cell.get_or_try_init(|| {
+        let actual_path = path.actual_path();
+        let (file, cell) = self.apk_cache.entry(actual_path)?;
+        let (apk, resolvers) = cell.get_or_try_init_(|| {
             let mmap = Mmap::builder()
                 .map(file)
-                .with_context(|| format!("failed to memory map `{}`", path.display()))?;
+                .with_context(|| format!("failed to memory map `{}`", actual_path.display()))?;
             let apk = zip::Archive::with_mmap(mmap)
-                .with_context(|| format!("failed to open zip file `{}`", path.display()))?;
+                .with_context(|| format!("failed to open zip file `{}`", actual_path.display()))?;
             let resolvers = InsertMap::new();
             Result::<_, Error>::Ok((apk, resolvers))
         })?;
@@ -850,28 +892,43 @@ impl Symbolizer {
     #[cfg(feature = "breakpad")]
     fn breakpad_resolver<'slf>(&'slf self, path: &Path) -> Result<&'slf BreakpadResolver> {
         let (file, cell) = self.breakpad_cache.entry(path)?;
-        let resolver = cell.get_or_try_init(|| self.create_breakpad_resolver(path, file))?;
+        let resolver = cell.get_or_try_init_(|| self.create_breakpad_resolver(path, file))?;
         Ok(resolver)
     }
 
-    fn create_perf_map(&self, path: &Path, file: &File) -> Result<PerfMap> {
+    fn create_perf_map_resolver(&self, path: &Path, file: &File) -> Result<PerfMap> {
         let perf_map = PerfMap::from_file(path, file)?;
         Ok(perf_map)
     }
 
-    fn perf_map(&self, pid: Pid) -> Result<Option<&PerfMap>> {
+    fn perf_map_resolver(&self, pid: Pid) -> Result<Option<&PerfMap>> {
         let path = PerfMap::path(pid);
 
         match self.perf_map_cache.entry(&path) {
             Ok((file, cell)) => {
-                let perf_map = cell.get_or_try_init(|| self.create_perf_map(&path, file))?;
+                let perf_map =
+                    cell.get_or_try_init_(|| self.create_perf_map_resolver(&path, file))?;
                 Ok(Some(perf_map))
             }
             Err(err) if err.kind() == ErrorKind::NotFound => Ok(None),
-            Err(err) => Err(err).with_context(|| format!("failed to open perf map `{path:?}`")),
+            Err(err) => {
+                Err(err).with_context(|| format!("failed to open perf map `{}`", path.display()))
+            }
         }
     }
 
+    fn vdso_parser<'slf>(
+        &'slf self,
+        pid: Pid,
+        range: &Range<Addr>,
+    ) -> Result<&'slf ElfParser<StaticMem>> {
+        let parser = self.vdso_parser.get_or_try_init_(|| {
+            let parser = create_vdso_parser(pid, range)?;
+            Result::<_, Error>::Ok(Box::new(parser))
+        })?;
+        Ok(parser)
+    }
+
     fn process_dispatch_resolver<'slf>(
         &'slf self,
         range: Range<Addr>,
@@ -903,13 +960,15 @@ impl Symbolizer {
         debug_syms: bool,
         perf_map: bool,
         map_files: bool,
-    ) -> Result<Vec<Symbolized>> {
+        vdso: bool,
+    ) -> Result<Vec<Symbolized<'_>>> {
         let mut handler = SymbolizeHandler {
             symbolizer: self,
             pid,
             debug_syms,
             perf_map,
             map_files,
+            vdso,
             all_symbols: Vec::with_capacity(addrs.len()),
         };
 
@@ -951,7 +1010,7 @@ impl Symbolizer {
 
     fn ksym_resolver<'slf>(&'slf self, path: &Path) -> Result<&'slf Rc<KsymResolver>> {
         let (file, cell) = self.ksym_cache.entry(path)?;
-        let resolver = cell.get_or_try_init(|| self.create_ksym_resolver(path, file))?;
+        let resolver = cell.get_or_try_init_(|| self.create_ksym_resolver(path, file))?;
         Ok(resolver)
     }
 
@@ -994,7 +1053,7 @@ impl Symbolizer {
             MaybeDefault::Some(vmlinux) => {
                 let resolver = self
                     .elf_cache
-                    .elf_resolver(vmlinux.as_path(), self.maybe_debug_dirs(*debug_syms))?;
+                    .elf_resolver(vmlinux, self.maybe_debug_dirs(*debug_syms))?;
                 Some(resolver)
             }
             MaybeDefault::Default => {
@@ -1012,7 +1071,7 @@ impl Symbolizer {
                 if let Some(vmlinux) = vmlinux {
                     let result = self
                         .elf_cache
-                        .elf_resolver(vmlinux.as_path(), self.maybe_debug_dirs(*debug_syms));
+                        .elf_resolver(&vmlinux, self.maybe_debug_dirs(*debug_syms));
                     match result {
                         Ok(resolver) => {
                             log::debug!("found suitable vmlinux file `{}`", vmlinux.display());
@@ -1068,7 +1127,7 @@ impl Symbolizer {
                 let _unpinned = self.elf_cache.unpin(path);
                 let result = self
                     .elf_cache
-                    .elf_resolver(path.as_path(), self.maybe_debug_dirs(false));
+                    .elf_resolver(path, self.maybe_debug_dirs(false));
                 // Make sure to always pin the entry, even if bailing
                 // due to a retrieval error. Basically, the semantics we
                 // want to have is that if caching new data fails the
@@ -1095,15 +1154,15 @@ impl Symbolizer {
     /// Symbolize a list of addresses.
     ///
     /// Symbolize a list of addresses using the provided symbolization
-    /// [`Source`][Source].
+    /// [`Source`].
     ///
     /// This function returns exactly one [`Symbolized`] object for each input
     /// address, in the order of input addresses.
     ///
     /// The following table lists which features the various formats
-    /// (represented by the [`Source`][Source] argument) support. If a feature
-    /// is not supported, the corresponding data in the [`Sym`] result will not
-    /// be populated.
+    /// (represented by the [`Source`] argument) support. If a feature is not
+    /// supported, the corresponding data in the [`Sym`] result will not be
+    /// populated.
     ///
     /// | Format      | Feature                          | Supported by format? | Supported by blazesym? |
     /// |-------------|----------------------------------|:--------------------:|:----------------------:|
@@ -1191,7 +1250,7 @@ impl Symbolizer {
             }) => {
                 let resolver = self
                     .elf_cache
-                    .elf_resolver(path.as_path(), self.maybe_debug_dirs(*debug_syms))?;
+                    .elf_resolver(path, self.maybe_debug_dirs(*debug_syms))?;
                 match input {
                     Input::VirtOffset(addrs) => addrs
                         .iter()
@@ -1242,6 +1301,7 @@ impl Symbolizer {
                 debug_syms,
                 perf_map,
                 map_files,
+                vdso,
                 _non_exhaustive: (),
             }) => {
                 let addrs = match input {
@@ -1258,7 +1318,7 @@ impl Symbolizer {
                     }
                 };
 
-                self.symbolize_user_addrs(addrs, *pid, *debug_syms, *perf_map, *map_files)
+                self.symbolize_user_addrs(addrs, *pid, *debug_syms, *perf_map, *map_files, *vdso)
             }
             #[cfg(feature = "gsym")]
             Source::Gsym(Gsym::Data(GsymData {
@@ -1375,7 +1435,7 @@ impl Symbolizer {
             }) => {
                 let resolver = self
                     .elf_cache
-                    .elf_resolver(path.as_path(), self.maybe_debug_dirs(*debug_syms))?;
+                    .elf_resolver(path, self.maybe_debug_dirs(*debug_syms))?;
                 let addr = match input {
                     Input::VirtOffset(addr) => addr,
                     Input::AbsAddr(..) => {
@@ -1416,6 +1476,7 @@ impl Symbolizer {
                 debug_syms,
                 perf_map,
                 map_files,
+                vdso,
                 _non_exhaustive: (),
             }) => {
                 let addr = match input {
@@ -1432,8 +1493,14 @@ impl Symbolizer {
                     }
                 };
 
-                let mut symbols =
-                    self.symbolize_user_addrs(&[addr], *pid, *debug_syms, *perf_map, *map_files)?;
+                let mut symbols = self.symbolize_user_addrs(
+                    &[addr],
+                    *pid,
+                    *debug_syms,
+                    *perf_map,
+                    *map_files,
+                    *vdso,
+                )?;
                 debug_assert!(symbols.len() == 1, "{symbols:#?}");
                 // SANITY: `symbolize_user_addrs` should *always* return
                 //         one result for one input (except on error
@@ -1509,6 +1576,9 @@ impl Default for Symbolizer {
 mod tests {
     use super::*;
 
+    use std::env::current_exe;
+
+    use test_fork::fork;
     use test_log::test;
 
     use crate::maps::Perm;
@@ -1539,6 +1609,15 @@ mod tests {
         });
     }
 
+    /// Test forcing a double check of all `Symbolizer` size changes.
+    #[cfg(target_pointer_width = "64")]
+    #[test]
+    fn symbolizer_size() {
+        // TODO: This size is rather large and we should look into
+        //       minimizing it.
+        assert_eq!(size_of::<Symbolizer>(), 1008);
+    }
+
     /// Check that we can correctly construct the source code path to a symbol.
     #[test]
     fn symbol_source_code_path() {
@@ -1559,11 +1638,11 @@ mod tests {
     #[test]
     fn demangle() {
         let symbol = Cow::Borrowed("_ZN4core9panicking9panic_fmt17h5f1a6fd39197ad62E");
-        let name = maybe_demangle(symbol, SrcLang::Rust);
+        let name = maybe_demangle_impl(symbol, SrcLang::Rust);
         assert_eq!(name, "core::panicking::panic_fmt");
 
         let symbol = Cow::Borrowed("_ZStlsISt11char_traitsIcEERSt13basic_ostreamIcT_ES5_PKc");
-        let name = maybe_demangle(symbol, SrcLang::Cpp);
+        let name = maybe_demangle_impl(symbol, SrcLang::Cpp);
         assert_eq!(
             name,
             "std::basic_ostream<char, std::char_traits<char> >& std::operator<< <std::char_traits<char> >(std::basic_ostream<char, std::char_traits<char> >&, char const*)"
@@ -1674,6 +1753,7 @@ mod tests {
             debug_syms: false,
             perf_map: false,
             map_files: false,
+            vdso: false,
             all_symbols: Vec::new(),
         };
         let () = normalize_sorted_user_addrs_with_entries(
@@ -1691,4 +1771,43 @@ mod tests {
             syms[0]
         );
     }
+
+    /// Check that we instantiate only a minimal number of resolvers
+    /// when using process symbolization with `map_files` (i.e., going
+    /// through symbolic links).
+    ///
+    /// Effectively, this is an integration test that makes sure that we
+    /// dereference symbolic links properly and not duplicate binary
+    /// parsing over and over, but it peeks at implementation details.
+    // Run in separate process to make sure that VMAs are not influenced
+    // by tests running concurrently.
+    #[fork]
+    #[test]
+    fn resolver_instantiation() {
+        let exe = current_exe().unwrap();
+        let addrs = maps::parse(Pid::Slf)
+            .unwrap()
+            .filter_map(|result| {
+                let entry = result.unwrap();
+                let path = entry.path_name.and_then(|path_name| {
+                    path_name.as_path().map(|path| path.symbolic_path.clone())
+                });
+                if path == Some(exe.clone()) {
+                    Some(entry.range.start)
+                } else {
+                    None
+                }
+            })
+            .collect::<Box<[_]>>();
+
+        assert!(addrs.len() > 1, "{:x?}", addrs.as_ref());
+
+        let src = Source::Process(Process::new(Pid::Slf));
+        let symbolizer = Symbolizer::new();
+        // We don't really care whether we could symbolize the addresses
+        // (unlikely), just that there was no error.
+        let _result = symbolizer.symbolize(&src, Input::AbsAddr(&addrs)).unwrap();
+
+        assert_eq!(symbolizer.elf_cache.entry_count(), 1);
+    }
 }
