@@ -1,3 +1,4 @@
+use std::cell::OnceCell;
 use std::cmp::Ordering;
 use std::ffi::CStr;
 use std::ffi::CString;
@@ -243,6 +244,28 @@ pub(crate) fn path_to_bytes(path: &Path) -> io::Result<&[u8]> {
     Ok(bytes)
 }
 
+// TODO: Remove once `OnceCell::get_or_try_init()` is stable.
+pub(crate) trait OnceCellExt<T> {
+    fn get_or_try_init_<F, E>(&self, f: F) -> Result<&T, E>
+    where
+        F: FnOnce() -> Result<T, E>;
+}
+
+impl<T> OnceCellExt<T> for OnceCell<T> {
+    fn get_or_try_init_<F, E>(&self, f: F) -> Result<&T, E>
+    where
+        F: FnOnce() -> Result<T, E>,
+    {
+        if let Some(value) = self.get() {
+            Ok(value)
+        } else {
+            let value = f()?;
+            Ok(self.get_or_init(|| value))
+        }
+    }
+}
+
+
 /// Reorder elements of `array` based on index information in `indices`.
 fn reorder<T, U>(array: &mut [T], indices: Vec<(U, usize)>) {
     debug_assert_eq!(array.len(), indices.len());
@@ -872,6 +895,7 @@ mod tests {
     /// Check that we can read various integers from a slice.
     #[tag(miri)]
     #[test]
+    #[expect(trivial_numeric_casts)]
     fn pod_reading() {
         macro_rules! test {
             ($type:ty) => {{
