@@ -14,6 +14,7 @@ use std::io::Write as _;
 use std::path::Path;
 use std::path::PathBuf;
 use std::process;
+use std::rc::Rc;
 
 use blazesym::helper::ElfResolver;
 use blazesym::inspect;
@@ -47,6 +48,9 @@ use blazesym::MaybeDefault;
 use blazesym::Mmap;
 use blazesym::Pid;
 use blazesym::Result;
+use blazesym::SymType;
+#[cfg(linux)]
+use blazesym::__private::find_gettimeofday_in_process;
 use blazesym::__private::find_the_answer_fn_in_zip;
 
 #[cfg(linux)]
@@ -58,7 +62,7 @@ use scopeguard::defer;
 use tempfile::tempdir;
 use tempfile::NamedTempFile;
 
-use test_fork::test as forked_test;
+use test_fork::fork;
 use test_log::test;
 use test_tag::tag;
 
@@ -199,7 +203,8 @@ fn symbolize_no_permission_impl(path: &Path) {
 /// Check that we fail symbolization as expected when we don't have the
 /// permission to open the symbolization source.
 #[cfg(linux)]
-#[forked_test]
+#[fork]
+#[test]
 fn symbolize_elf_no_permission() {
     use libc::getresuid;
     use std::os::unix::fs::PermissionsExt as _;
@@ -235,6 +240,40 @@ fn symbolize_elf_no_permission() {
     as_user(ruid, uid, || symbolize_no_permission_impl(path))
 }
 
+/// Check that we correctly symbolize untyped symbols.
+#[tag(other_os)]
+#[test]
+fn symbolize_elf_no_type() {
+    let path = Path::new(&env!("CARGO_MANIFEST_DIR"))
+        .join("data")
+        .join("test-stable-addrs.bin");
+    let src = inspect::source::Source::Elf(inspect::source::Elf::new(path));
+    let inspector = inspect::Inspector::new();
+    let results = inspector
+        .lookup(&src, &["untyped"])
+        .unwrap()
+        .into_iter()
+        .flatten()
+        .collect::<Vec<_>>();
+    assert_eq!(results.len(), 1);
+    let untyped = &results[0];
+    assert_eq!(untyped.sym_type, SymType::Undefined);
+
+    let path = Path::new(&env!("CARGO_MANIFEST_DIR"))
+        .join("data")
+        .join("test-stable-addrs.bin");
+    let src = Source::from(Elf::new(path));
+    let symbolizer = Symbolizer::new();
+    let result = symbolizer
+        .symbolize_single(&src, Input::VirtOffset(untyped.addr))
+        .unwrap()
+        .into_sym()
+        .unwrap();
+
+    assert_eq!(result.name, "untyped");
+    assert_eq!(result.addr, untyped.addr);
+}
+
 /// Check that we correctly symbolize zero sized symbols.
 // TODO: Extend this test to more formats.
 #[tag(other_os)]
@@ -524,6 +563,7 @@ fn symbolize_dwarf_gsym_inlined() {
     for file in [
         "test-stable-addrs-stripped-elf-with-dwarf.bin",
         "test-stable-addrs-stripped-with-link.bin",
+        "test-stable-addrs-split-dwarf.bin",
         "test-stable-addrs-compressed-debug-zlib.bin",
         #[cfg(feature = "zstd")]
         "test-stable-addrs-compressed-debug-zstd.bin",
@@ -575,6 +615,24 @@ fn symbolize_dwarf_non_existent_debug_link() {
     assert_eq!(result, None);
 }
 
+/// Check that we don't error out due to checksum mismatch on
+/// self-referential debug link.
+#[test]
+fn symbolize_dwarf_self_referential_debug_link() {
+    let path = Path::new(&env!("CARGO_MANIFEST_DIR"))
+        .join("data")
+        .join("test-stable-addrs-with-link-to-self.bin");
+    let src = Source::from(Elf::new(path));
+    let symbolizer = Symbolizer::builder().enable_auto_reload(false).build();
+    let result = symbolizer
+        .symbolize_single(&src, Input::VirtOffset(0x2000200))
+        .unwrap()
+        .into_sym()
+        .unwrap();
+    assert_eq!(result.name, "factorial");
+    assert_eq!(result.addr, 0x2000200);
+}
+
 /// Check that we honor configured debug directories as one would expect.
 #[tag(other_os)]
 #[test]
@@ -764,6 +822,65 @@ fn symbolize_dwarf_demangle() {
     let elf = inspect::source::Elf::new(&test_dwarf);
     let src = inspect::source::Source::Elf(elf);
 
+    let inspector = inspect::Inspector::new();
+    let results = inspector
+        .lookup(&src, &["_RNvCs69hjMPjVIJK_4test13test_function"])
+        .unwrap()
+        .into_iter()
+        .flatten()
+        .collect::<Vec<_>>();
+    assert_eq!(results.len(), 1);
+
+    let addr = results[0].addr;
+    let size = results[0].size.unwrap() as u64;
+    for inst_addr in addr..addr + size {
+        if test(&test_dwarf, inst_addr).is_ok() {
+            return
+        }
+    }
+
+    panic!("failed to find inlined function call");
+}
+
+/// Check that we can symbolize an address with inline function
+/// information inside a DWARF package.
+#[test]
+fn symbolize_rust_dwp() {
+    fn test(test_dwarf: &Path, addr: Addr) -> Result<(), ()> {
+        let src = Source::Elf(Elf::new(test_dwarf));
+        let symbolizer = Symbolizer::new();
+        let sym = symbolizer
+            .symbolize_single(&src, Input::VirtOffset(addr))
+            .unwrap()
+            .into_sym()
+            .unwrap();
+
+        assert_eq!(sym.name, "test::test_function");
+
+        let code_info = sym.code_info.as_ref().unwrap();
+        assert_ne!(code_info.dir, None);
+        assert_eq!(code_info.file, OsStr::new("test.rs"));
+        assert!(code_info.line.is_some());
+
+        if sym.inlined.is_empty() {
+            return Err(())
+        }
+        assert_eq!(sym.inlined.len(), 1, "{:#?}", sym.inlined);
+        let inlined = &sym.inlined[0];
+        assert_eq!(inlined.name, "test::inlined_call");
+        let code_info = inlined.code_info.as_ref().unwrap();
+        assert_ne!(code_info.dir, None);
+        assert_eq!(code_info.file, OsStr::new("test.rs"));
+        assert!(code_info.line.is_some());
+        Ok(())
+    }
+
+    let test_dwarf = Path::new(&env!("CARGO_MANIFEST_DIR"))
+        .join("data")
+        .join("test-rs-split-dwarf.bin");
+    let elf = inspect::source::Elf::new(&test_dwarf);
+    let src = inspect::source::Source::Elf(elf);
+
     let inspector = inspect::Inspector::new();
     let results = inspector
         .lookup(&src, &["_RNvCs69hjMPjVIJK_4test13test_function"])
@@ -774,16 +891,7 @@ fn symbolize_dwarf_demangle() {
     assert!(!results.is_empty());
 
     let addr = results[0].addr;
-    let src = Source::Elf(Elf::new(&test_dwarf));
-    let symbolizer = Symbolizer::builder().enable_demangling(false).build();
-    let result = symbolizer
-        .symbolize_single(&src, Input::VirtOffset(addr))
-        .unwrap()
-        .into_sym()
-        .unwrap();
-
-    let addr = result.addr;
-    let size = result.size.unwrap() as u64;
+    let size = results[0].size.unwrap() as u64;
     for inst_addr in addr..addr + size {
         if test(&test_dwarf, inst_addr).is_ok() {
             return
@@ -910,34 +1018,37 @@ fn symbolize_process_exited_cached_vmas() {
 /// Check that we can symbolize an address residing in a zip archive.
 #[test]
 fn symbolize_process_zip() {
-    let test_zip = Path::new(&env!("CARGO_MANIFEST_DIR"))
-        .join("data")
-        .join("test.zip");
+    fn test(map_files: bool) {
+        let test_zip = Path::new(&env!("CARGO_MANIFEST_DIR"))
+            .join("data")
+            .join("test.zip");
 
-    let mmap = Mmap::builder().exec().open(test_zip).unwrap();
-    let (sym, the_answer_addr) = find_the_answer_fn_in_zip(&mmap);
+        let mmap = Mmap::builder().exec().open(&test_zip).unwrap();
+        let (sym, the_answer_addr) = find_the_answer_fn_in_zip(&mmap);
 
-    // Symbolize the address we just looked up. It should be correctly
-    // mapped to the `the_answer` function within our process.
-    let src = Source::Process(Process::new(Pid::Slf));
-    let symbolizer = Symbolizer::new();
-    let result = symbolizer
-        .symbolize_single(&src, Input::AbsAddr(the_answer_addr))
-        .unwrap()
-        .into_sym()
-        .unwrap();
+        // Symbolize the address we just looked up. It should be correctly
+        // mapped to the `the_answer` function within our process.
+        let mut process = Process::new(Pid::Slf);
+        process.map_files = map_files;
+        let src = Source::Process(process);
+        let symbolizer = Symbolizer::new();
+        let result = symbolizer
+            .symbolize_single(&src, Input::AbsAddr(the_answer_addr))
+            .unwrap()
+            .into_sym()
+            .unwrap();
 
-    assert_eq!(result.name, "the_answer");
-    assert!(
-        result
-            .module
-            .as_deref()
-            .map(|module| module.to_string_lossy().ends_with("!/libtest-so.so"))
-            .unwrap_or(false),
-        "{:?}",
-        result.module
-    );
-    assert_eq!(result.addr, sym.addr);
+        let mut module = test_zip.as_os_str().to_os_string();
+        let () = module.push("!/libtest-so.so");
+
+        assert_eq!(result.name, "the_answer");
+        assert_eq!(result.module.as_deref(), Some(module.as_os_str()));
+        assert_eq!(result.addr, sym.addr);
+    }
+
+    for map_files in [false, true] {
+        let () = test(map_files);
+    }
 }
 
 /// Test that we can use a custom dispatch function when symbolizing addresses
@@ -988,10 +1099,63 @@ fn symbolize_process_with_custom_dispatch() {
     test(process_no_dispatch);
 }
 
+/// Test that we symbolize addresses in a vDSO in our process.
+#[cfg(linux)]
+#[test]
+fn symbolize_own_process_vdso() {
+    use libc::clock_gettime;
+    use libc::gettimeofday;
+
+    let src = Source::Process(Process::new(Pid::Slf));
+    // Both functions are typically provided by the vDSO, though there
+    // is no guarantee of that.
+    let addrs = [gettimeofday as Addr, clock_gettime as Addr];
+    let symbolizer = Symbolizer::new();
+
+    // Symbolize twice, to exercise both cache population and cache
+    // usage paths.
+    for _ in [0, 1] {
+        let results = symbolizer
+            .symbolize(&src, Input::AbsAddr(&addrs))
+            .unwrap()
+            .into_iter()
+            .collect::<Vec<_>>();
+        assert_eq!(results.len(), 2);
+        // Given that we can't guarantee that these symbols are in a vDSO,
+        // it's hard for us to assert anything more than the name.
+        let sym1 = results[0].as_sym().unwrap();
+        assert!(sym1.name.ends_with("gettimeofday"), "{sym1:?}");
+        let sym2 = results[1].as_sym().unwrap();
+        assert!(sym2.name.contains("clock_gettime"), "{sym2:?}");
+    }
+}
+
+/// Test that we symbolize addresses in a vDSO in a "remote" process.
+#[cfg(linux)]
+#[test]
+fn symbolize_remote_process_vdso() {
+    let test_block = Path::new(&env!("CARGO_MANIFEST_DIR"))
+        .join("data")
+        .join("test-block.bin");
+    let () = RemoteProcess::default().exec(&test_block, |pid, _addr| {
+        let addr = find_gettimeofday_in_process(pid);
+
+        let process = Process::new(pid);
+        let src = Source::Process(process);
+        let symbolizer = Symbolizer::new();
+        let sym = symbolizer
+            .symbolize_single(&src, Input::AbsAddr(addr))
+            .unwrap()
+            .into_sym()
+            .unwrap();
+        assert!(sym.name.ends_with("gettimeofday"), "{sym:?}");
+    });
+}
 
 /// Make sure that we do not fail symbolization when an empty perf
 /// map is present.
-#[forked_test]
+#[fork]
+#[test]
 fn symbolize_with_empty_perf_map() {
     let heap = vec![0; 4096];
     let path = format!("/tmp/perf-{}.map", process::id());
@@ -1019,7 +1183,8 @@ fn symbolize_with_empty_perf_map() {
 
 /// Check that we can symbolize an address using a perf map.
 #[cfg(linux)]
-#[forked_test]
+#[fork]
+#[test]
 #[ignore = "test requires python 3.12 or higher"]
 fn symbolize_process_perf_map() {
     use std::ffi::OsString;
@@ -1103,7 +1268,8 @@ fn symbolize_permissionless_impl(pid: Pid, addr: Addr, _test_lib: &Path) {
 /// Check that we can symbolize an address in a process using only
 /// symbolic paths.
 #[cfg(linux)]
-#[forked_test]
+#[fork]
+#[test]
 fn symbolize_process_symbolic_paths() {
     run_unprivileged_process_test(symbolize_permissionless_impl)
 }
@@ -1139,7 +1305,7 @@ fn symbolize_zip_with_custom_dispatch() {
         let src = Source::Process(Process::new(Pid::Slf));
         let symbolizer = Symbolizer::builder().set_apk_dispatcher(dispatcher).build();
         let result = symbolizer
-            .symbolize_single(&src, Input::AbsAddr(the_answer_addr as Addr))
+            .symbolize_single(&src, Input::AbsAddr(the_answer_addr))
             .unwrap()
             .into_sym()
             .unwrap();
@@ -1200,7 +1366,7 @@ fn symbolize_zip_with_custom_dispatch_errors() {
         let src = Source::Process(Process::new(Pid::Slf));
         let symbolizer = Symbolizer::builder().set_apk_dispatcher(dispatcher).build();
         let err = symbolizer
-            .symbolize_single(&src, Input::AbsAddr(the_answer_addr as Addr))
+            .symbolize_single(&src, Input::AbsAddr(the_answer_addr))
             .unwrap_err();
 
         assert_eq!(err.to_string(), "induced error");
@@ -1277,12 +1443,16 @@ fn symbolize_kernel_kallsyms() {
         };
         let src = Source::Kernel(kernel);
         let symbolizer = Symbolizer::new();
-        let symbolized = symbolizer
-            .symbolize_single(&src, Input::AbsAddr(0xc080a470))
-            .unwrap();
-        let init_task_sym = symbolized.into_sym().unwrap();
-        assert_eq!(init_task_sym.name, "init_task");
-        assert_eq!(init_task_sym.code_info, None);
+
+        for offset in 0u16..20 {
+            let symbolized = symbolizer
+                .symbolize_single(&src, Input::AbsAddr(0xc080a470 + Addr::from(offset)))
+                .unwrap();
+            let init_task_sym = symbolized.into_sym().unwrap();
+            assert_eq!(init_task_sym.name, "init_task");
+            assert_eq!(init_task_sym.code_info, None);
+            assert_eq!(init_task_sym.offset, usize::from(offset));
+        }
     }
 
     test(MaybeDefault::None);
@@ -1365,15 +1535,19 @@ fn symbolize_kernel_system_vmlinux() {
         let mut pairs = content
             .lines()
             .filter_map(|line| {
-                let [addr, ty, name] = line
-                    .split_ascii_whitespace()
-                    .collect::<Vec<_>>()
-                    .get(0..3)?
-                    .try_into()
-                    .unwrap();
+                let fields = line.split_ascii_whitespace().collect::<Vec<_>>();
+                let (addr, ty, name, module) = match fields[..] {
+                    [addr, ty, name] => (addr, ty, name, None),
+                    [addr, ty, name, module] => (addr, ty, name, Some(module)),
+                    _ => panic!("encountered unexpected kallsyms line: {line}"),
+                };
                 if !["D", "T", "t", "W"].contains(&ty) {
                     return None
                 }
+                // TODO: Eventually we need to support modules.
+                if module.is_some() {
+                    return None
+                }
                 let addr = Addr::from_str_radix(addr, 16).unwrap();
                 Some((addr, name))
             })
@@ -1434,7 +1608,18 @@ fn symbolize_kernel_system_vmlinux() {
         let sym = sym
             .as_sym()
             .unwrap_or_else(|| panic!("failed to symbolize {:x?}", syms[i]));
-        assert_eq!(sym.name, syms[i].1, "{sym:?} | {:?}", syms[i]);
+        // We have seen cases where symbols were suffixed with LLVM
+        // specific prefixes in kallsyms but not in the ELF file. But,
+        // in all likelihood, that can happen the other way around as
+        // well..
+        // E.g.,
+        //   left: "fanotify_free_mark"
+        //   right: "fanotify_free_mark.llvm.12215866716532162847"
+        assert!(
+            sym.name.starts_with(&syms[i].1) || syms[i].1.starts_with(sym.name.as_ref()),
+            "{sym:?} | {:?}",
+            syms[i]
+        );
     }
 }
 
@@ -1456,7 +1641,11 @@ fn symbolize_kernel_bpf_program() {
             .symbolize(&src, Input::AbsAddr(&[handle_getpid, subprogram]))
             .unwrap();
         let handle_getpid_sym = result[0].as_sym().unwrap();
-        assert_eq!(handle_getpid_sym.name, "handle__getpid");
+        assert!(
+            handle_getpid_sym.name.ends_with("handle__getpid"),
+            "{}",
+            handle_getpid_sym.name
+        );
         let code_info = handle_getpid_sym.code_info.as_ref().unwrap();
         assert_eq!(code_info.dir, None);
         assert_eq!(
@@ -1467,7 +1656,11 @@ fn symbolize_kernel_bpf_program() {
         assert_ne!(code_info.column, None);
 
         let subprogram_sym = result[1].as_sym().unwrap();
-        assert_eq!(subprogram_sym.name, "subprogram");
+        assert!(
+            subprogram_sym.name.ends_with("subprogram"),
+            "{}",
+            subprogram_sym.name
+        );
         let code_info = subprogram_sym.code_info.as_ref().unwrap();
         assert_eq!(code_info.dir, None);
         assert_eq!(
@@ -1514,3 +1707,34 @@ fn symbolize_normalized_large_memsize() {
         assert_eq!(sym.name, "_start");
     });
 }
+
+/// Make sure that registering an [`ElfResolver`] with a [`Symbolizer`]
+/// fails if one is already present for a given path.
+#[test]
+fn register_an_existing_elfresolver() {
+    let bin_name = Path::new(&env!("CARGO_MANIFEST_DIR"))
+        .join("data")
+        .join("test-stable-addrs.bin");
+
+    let resolver = Rc::new(ElfResolver::open(&bin_name).unwrap());
+    let mut symbolizer = Symbolizer::new();
+    let () = symbolizer
+        .register_elf_resolver(&bin_name, Rc::clone(&resolver))
+        .unwrap();
+
+    let err = symbolizer
+        .register_elf_resolver(&bin_name, Rc::clone(&resolver))
+        .unwrap_err();
+
+    assert_eq!(err.kind(), ErrorKind::AlreadyExists);
+}
+
+/// Make sure that creating an [`ElfResolver`] from a non-existing file
+/// fails with an actual error instead of panicking.
+#[test]
+fn create_elf_resolver_from_non_existing_path() {
+    let path = Path::new("/This/Path/Does.Not/Exist");
+    let err = ElfResolver::open(path).unwrap_err();
+
+    assert_eq!(err.kind(), ErrorKind::NotFound);
+}
